# -*- coding: utf-8 -*-
"""
Created on Sat Dec  7 18:32:34 2019

@author: Angshuman Deb
"""

import numpy as np
from pyDOE2 import lhs
import sys
from scipy import integrate
from scipy import interpolate


class Mixture:

    def __init__(self, name):
        self.name = name

    def __call__(self, prob_dists, **kwargs):
        return FrozenMixture(self, prob_dists, **kwargs)

    def mean(self, prob_dists, **kwargs):
        return FrozenMixture(self, prob_dists, **kwargs).mean()

    def std(self, prob_dists, **kwargs):
        return FrozenMixture(self, prob_dists, **kwargs).std()

    def support(self, prob_dists, **kwargs):
        return FrozenMixture(self, prob_dists, **kwargs).support()

    def pdf(self, x, prob_dists, **kwargs):
        return FrozenMixture(self, prob_dists, **kwargs).pdf(x)

    def cdf(self, x, prob_dists, **kwargs):
        return FrozenMixture(self, prob_dists, **kwargs).cdf(x)

    def ppf(self, x, prob_dists, **kwargs):
        return FrozenMixture(self, prob_dists, **kwargs).ppf(x)

    def entropy(self, prob_dists, **kwargs):
        return FrozenMixture(self, prob_dists, **kwargs).entropy()

    def rvs(self, prob_dists, **kwargs):
        size = kwargs.get('size', None)
        random_state = kwargs.get('random_state', None)
        method = kwargs.get('method', 'mcs')
        each = kwargs.get('each', 'arbitrary')
        weights = kwargs.get('weights', np.array([1.0 / len(prob_dists)] * len(prob_dists)))
        return FrozenMixture(self, prob_dists, each=each,
                             weights=weights).rvs(size=size, random_state=random_state, method=method)


mixture = Mixture('mixture')


# Frozen Mixture
class FrozenMixture:

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, dist, prob_dists, **kwargs):
        # prob_dists : either a 2d array or a list

        self.dist = dist
        self.each = kwargs.get('each', 'arbitrary')
        self.mixture_size = len(prob_dists)
        self.weights = kwargs.get('weights', np.array([1.0 / self.mixture_size] * self.mixture_size))

        if self.each == 'arbitrary':
            # prob_dists is a list
            self.prob_dists = prob_dists
            self.mean_each = np.array([prob_dist.mean() for prob_dist in self.prob_dists])
            self.std_each = np.array([prob_dist.std() for prob_dist in self.prob_dists])
            self.support_each = np.array([prob_dist.support() for prob_dist in self.prob_dists])
            self.args = [FrozenMixture.get_all_dist_params(prob_dist) for prob_dist in self.prob_dists]
            self.kwds = dict()
        else:
            # prob_dists is a 2d array
            self.prob_dists = self.each(*prob_dists.T)
            self.mean_each = self.prob_dists.mean()
            self.std_each = self.prob_dists.std()
            self.support_each = np.array(self.prob_dists.support()).T
            self.args = [list(item) for item in prob_dists]
            self.kwds = dict()

        self._mean = np.sum(self.mean_each * self.weights)
        self._std = np.sum(
            self.weights * (np.square(self.std_each) + np.square(self.mean_each - self._mean))) ** 0.5
        self._support = (np.min(self.support_each[:, 0]), np.max(self.support_each[:, 1]))
        self._k = 8
        self._approx_support = FrozenMixture.get_approximate_support(self._support, self._mean, self._std, self._k)
        self._ppf_data = None

    ####################################################################################################################
    # Public functionalities
    ####################################################################################################################

    def mean(self):
        return self._mean

    # ------------------------------------------------------------------------------------------------------------------

    def std(self):
        return self._std

    # ------------------------------------------------------------------------------------------------------------------

    def median(self):
        return self.ppf(0.5)

    # ------------------------------------------------------------------------------------------------------------------

    def support(self):
        return self._support

    # ------------------------------------------------------------------------------------------------------------------

    def pdf(self, x):
        scalar = True if np.isscalar(x) else False
        x = np.array(x)
        x_temp = x.ravel()[:, np.newaxis]
        temp = self.weights
        for itr in range(len(x_temp.shape)):
            temp = temp[:, np.newaxis]

        if self.each == 'arbitrary':
            pdf_x = list()
            for itr in range(self.mixture_size):
                pdf_x.append(self.prob_dists[itr].pdf(x_temp))
            pdf_x = np.array(pdf_x)
        else:
            pdf_x_temp = self.prob_dists.pdf(x_temp)
            pdf_x = np.einsum('ijd->jid', pdf_x_temp[:, :, np.newaxis])

        pdf_x = pdf_x * temp
        pdf_x = np.sum(pdf_x, axis=0).reshape(x.shape)
        return pdf_x.item() if scalar else pdf_x

    # ------------------------------------------------------------------------------------------------------------------

    def cdf(self, x):
        scalar = True if np.isscalar(x) else False
        x = np.array(x)
        x_temp = x.ravel()[:, np.newaxis]
        temp = self.weights
        for itr in range(len(x_temp.shape)):
            temp = temp[:, np.newaxis]

        if self.each == 'arbitrary':
            cdf_x = list()
            for itr in range(self.mixture_size):
                cdf_x.append(self.prob_dists[itr].cdf(x_temp))
            cdf_x = np.array(cdf_x)
        else:
            cdf_x_temp = self.prob_dists.cdf(x_temp)
            cdf_x = np.einsum('ijd->jid', cdf_x_temp[:, :, np.newaxis])

        cdf_x = cdf_x * temp
        cdf_x = np.sum(cdf_x, axis=0).reshape(x.shape)
        return cdf_x.item() if scalar else cdf_x

    # ------------------------------------------------------------------------------------------------------------------

    def entropy(self):
        eps = sys.float_info.epsilon
        entr = integrate.quad(FrozenMixture.entropy_integrand,
                              self._approx_support[0] + eps,
                              self._approx_support[-1] - eps,
                              args=(self,))
        return entr[0]

    # ------------------------------------------------------------------------------------------------------------------

    # def rvs(self, size=None, random_state=None):
    #     np.random.seed(random_state)
    #     mixture_inds = np.random.choice(self.mixture_size, size=size, p=self.weights)
    #     if np.isscalar(mixture_inds):
    #         mixture_inds = np.array([mixture_inds])
    #     mixture_inds = mixture_inds.flatten()
    #     if self.each == 'arbitrary':
    #         rvs_temp = np.array([self.prob_dists[ind].rvs() for ind in mixture_inds])
    #     else:
    #         rvs_temp = np.array([self.prob_dists.rvs()[ind] for ind in mixture_inds])
    #     return rvs_temp.reshape(size) if size is not None else rvs_temp.item()

    # ------------------------------------------------------------------------------------------------------------------

    def rvs(self, size=None, random_state=None, method='mcs'):
        if size is None:
            size = 1
            none = True
        else:
            none = False

        x = np.zeros(size)
        if method == 'mcs':
            np.random.seed(random_state)
            x = np.random.uniform(0.0, 1.0, size)
        elif method == 'lhs':
            if not np.isscalar(size):
                n_samples = size[0]
                if len(size) == 2:
                    n_set = size[1]
                elif len(size) > 2:
                    raise ValueError("Non-scalar size for this type of sampling method "
                                     "should not be of length more than 2 and should represent "
                                     "(number of realizations of random variable, number of sets of realizations)!")
                else:
                    n_set = 1
            else:
                n_samples = size
                n_set = 1

            for itr_set in range(n_set):
                if random_state is None:
                    rng_seed = None
                else:
                    if itr_set > 0:
                        rng_seed = int(''.join([str(i) for i in [random_state, itr_set]]))
                    else:
                        rng_seed = random_state
                temp = lhs(1, n_samples, random_state=rng_seed)
                if len(x.shape) == 2:
                    x[:, itr_set] = temp
                else:
                    x = temp

        rvs_temp = self.ppf(x)
        return rvs_temp.reshape(size) if not none else rvs_temp.item()

    # ------------------------------------------------------------------------------------------------------------------

    def ppf(self, p):
        if self._ppf_data is None:
            self.set_ppf_data()

        ppf_interp = interpolate.interp1d(self._ppf_data[:, 1], self._ppf_data[:, 0], fill_value="extrapolate")
        x = ppf_interp(p)

        if not np.isscalar(p):
            scalar = False
            x[p < 0.0] = np.nan
            x[p > 1.0] = np.nan
            x[p == 0.0] = self._support[0]
            x[p == 1.0] = self._support[1]
        else:
            scalar = True
            if p < 0.0:
                x = np.nan
            elif p > 1.0:
                x = np.nan
            elif p == 0.0:
                x = self._support[0]
            elif p == 1.0:
                x = self._support[1]
        return x.item() if scalar else x

    # ------------------------------------------------------------------------------------------------------------------

    def set_ppf_data(self):
        # p = np.linspace(self.cdf(self._approx_support[0]), self.cdf(self._approx_support[-1]), 500)
        # x = self.ppf_helper_bisection(p,
        #                               np.full(len(p), self._approx_support[0]),
        #                               np.full(len(p), self._approx_support[-1]))

        x = np.linspace(*self._approx_support, 1000)
        p = self.cdf(x)
        self._ppf_data = np.column_stack((x, p))

    # ------------------------------------------------------------------------------------------------------------------

    def ppf_helper_bisection(self, p, a, b, max_iterations=100, tolerance=1e-4):

        if p.size == 0:
            return np.empty(0)

        f_a = self.cdf(a) - p

        for i in range(max_iterations):

            x = a + (b - a) / 2
            f_x = self.cdf(x) - p

            if np.sum(np.abs(f_x) < tolerance) == len(f_x):
                return x

            ind1 = np.sign(f_a) * np.sign(f_x) > 0
            ind2 = ~ind1

            a[ind1] = x[ind1]
            f_a[ind1] = f_x[ind1]
            b[ind2] = x[ind2]

        return a + (b - a) / 2

    ####################################################################################################################
    # Static methods
    ####################################################################################################################

    @staticmethod
    def entropy_integrand(x, m):
        px = m.pdf(x)
        return -px * np.log(px)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_approximate_support(supp, m, s, k):
        return max(m - k * s, supp[0]), min(m + k * s, supp[-1])

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_all_dist_params(prob_dist):
        temp = prob_dist.dist.__dict__['_parse_args'](*prob_dist.args, **prob_dist.kwds)
        return [*temp[0], *temp[1:len(temp)]]

    # ------------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 15 11:14:02 2019

@author: Angshuman Deb
"""

from .structure import Structure
from .utility import Utility
from abc import ABC, abstractmethod
import numpy as np
import os
import scipy.io as sio
import matplotlib.pyplot as plt


class IM(ABC):

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, structure, gmm):
        self.structure = structure
        self.gmm = gmm

    ####################################################################################################################
    # Abstract methods
    ####################################################################################################################

    @abstractmethod
    def evaluate_gmm(self, *args, **kwargs):
        pass

    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def select_ground_motion_records(self, *args, **kwargs):
        pass

    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def compute_seismic_hazard_integral(self, *args, **kwargs):
        pass

    ####################################################################################################################
    # Public functionalities (instance methods)
    ####################################################################################################################

    def _compute_seismic_hazard_integral(self, prob_dist_im, prob_dist_params_im, im_input):
        im_input = im_input.reshape(len(im_input), 1)  # 2-d vector
        structure = self.structure
        site_hazard_info = structure.get_site_hazard_info()
        scenario_rate = site_hazard_info['scenario_rate']  # 1-d array
        prob_dist_im = prob_dist_im(*prob_dist_params_im)
        nu_im_scenario = (1 - prob_dist_im.cdf(im_input)) * scenario_rate
        nu_im = np.sum(nu_im_scenario, axis=1)  # sum along (over) scenarios

        to_return = {'seismic_hazard_curve': np.column_stack((im_input, nu_im)),
                     'seismic_hazard_curve_deagg': nu_im_scenario}

        return to_return

    # ------------------------------------------------------------------------------------------------------------------

    def get_inv_seismic_hazard(self, mrp, for_which):
        shc, _ = self.get_psha_results(for_which)
        return shc[np.argmin(np.abs(shc[:, 1] - 1 / mrp)), 0]

    # ------------------------------------------------------------------------------------------------------------------

    def get_seismic_hazard_deagg(self, mrp, for_which):
        shc, shc_deagg = self.get_psha_results(for_which)
        ind = np.argmin(np.abs(shc[:, 1] - 1 / mrp))
        # return 1-d array
        # return shc_deagg[ind, :] / shc[ind, 1]
        return (shc_deagg[ind, :] - shc_deagg[ind + 1, :]) / (shc[ind, 1] - shc[ind + 1, 1])

    # ------------------------------------------------------------------------------------------------------------------

    def get_psha_results(self, for_which):
        model_work_dir_path = self.structure.model_work_dir_path
        dir_category = 'PSHA_Results'
        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)
        file_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category,
                                     *Structure.get_modified_for_which(for_which, 'Design_Num'), 'psha_results.pickle')
        run_case_str = ' '.join(for_which)
        psha_results = Utility.pickle_load_dict(file_path)[run_case_str]
        return psha_results['seismic_hazard_curve'], psha_results['seismic_hazard_curve_deagg']

    # ------------------------------------------------------------------------------------------------------------------

    def get_gms_results(self, for_which):
        model_work_dir_path = self.structure.model_work_dir_path
        dir_category = 'GMS_Results'
        dir_level_to_seek = 'Hazard_Level_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)
        file_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category,
                                     *Structure.get_modified_for_which(for_which, 'Design_Num'), 'gms_results.pickle')
        run_case_str = ' '.join(for_which)
        gms_results = Utility.pickle_load_dict(file_path)[run_case_str]

        to_return = list()
        to_return.append(gms_results['ground_motion_records'])
        to_return.append(gms_results['target_spectra'])
        to_return.append(gms_results['mrp'])
        to_return.append(gms_results['n_gm'])

        return to_return

    # ------------------------------------------------------------------------------------------------------------------

    def get_gm_rec_info(self, for_which):
        dir_level_to_seek = 'Ground_Motion_Rec_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        gm_records, _, _, _ = self.get_gms_results(for_which)
        gm_index = int(for_which[Structure.get_dir_level_index('Ground_Motion_Rec_Num')]) - 1
        gm_file_names = [gm_records[key][gm_index] for key in gm_records.keys() if 'file_name_' in key]
        scale_fac = gm_records['scale_factor'][gm_index]

        return gm_file_names, scale_fac

    # ------------------------------------------------------------------------------------------------------------------

    def plot_shc(self, for_which, **kwargs):
        figkwargs = kwargs.get('figkwargs', dict())
        grid_alpha = kwargs.get('grid_alpha', 0.5)
        minor_grid_alpha = kwargs.get('minor_grid_alpha', 0.0)
        save_dir_path = kwargs.get('save_dir_path', Utility.get_path(os.getcwd()))
        save_mat = kwargs.get('save_mat', False)
        lc = kwargs.get('lc', 'black')
        lw = kwargs.get('lw', 1.5)
        fig_ax = kwargs.get('fig_ax', None)

        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig = plt.figure(**figkwargs)  # also accepts tight_layout=bool
            ax = fig.gca()
        else:
            fig = fig_ax[0]
            ax = fig_ax[1]
        ax.minorticks_on()
        ax.grid(True, which="major", alpha=grid_alpha)
        ax.grid(True, which="minor", alpha=minor_grid_alpha)

        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        shc = self.get_psha_results(for_which)[0]
        ax.loglog(shc[:, 0], shc[:, 1], '-', color=lc, linewidth=lw)

        export_mat_dict = dict()
        export_mat_dict['shc'] = shc

        if save_mat:
            structure = self.structure
            name = structure.name
            for_which_str = '_'.join(for_which)
            file_name = f'plot_shc_{name}_{for_which_str}.mat'
            sio.savemat(Utility.get_path(save_dir_path, file_name), export_mat_dict)

        return fig, ax, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------

    def plot_seismic_hazard_deagg(self, mrp, for_which, **kwargs):
        figkwargs = kwargs.get('figkwargs', dict())
        cmap = kwargs.get('cmap', 'viridis')
        grid_alpha = kwargs.get('grid_alpha', 0.5)
        minor_grid = kwargs.get('minor_grid', False)
        save_dir_path = kwargs.get('save_dir_path', Utility.get_path(os.getcwd()))
        save_mat = kwargs.get('save_mat', False)
        fig_ax = kwargs.get('fig_ax', None)

        structure = self.structure
        site_hazard_info = structure.get_site_hazard_info()
        mag = site_hazard_info['scenario_data']['mag']
        dist = site_hazard_info['scenario_data']['dist']
        rate = self.get_seismic_hazard_deagg(mrp, for_which)

        fig, ax, export_mat_dict = Utility.plot_3d_bar(mag, dist, rate,
                                                       figkwargs=figkwargs, cmap=cmap,
                                                       grid_alpha=grid_alpha, minor_grid=minor_grid,
                                                       fig_ax=fig_ax)

        export_mat_dict['mag'] = export_mat_dict.pop('x')
        export_mat_dict['dist'] = export_mat_dict.pop('y')
        export_mat_dict['rate'] = export_mat_dict.pop('dz')

        if save_mat:
            name = structure.name
            for_which_str = '_'.join(for_which)
            file_name = f'plot_seismic_hazard_deagg_{name}_{for_which_str}_mrp_{mrp}.mat'
            sio.savemat(Utility.get_path(save_dir_path, file_name), export_mat_dict)

        return fig, ax, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------

    def get_gmm_param_names(self):
        gmm = self.gmm
        params = gmm.PARAMS
        reqd_param_names = [param.name for param in params]
        return reqd_param_names

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def invert_seismic_hazard(shc, mrp):
        return shc[np.argmin(np.abs(shc[:, 1] - 1 / mrp)), 0]



# -*- coding: utf-8 -*-
"""
Created on Sun Nov 17 16:16:25 2019

@author: Angshuman Deb
"""

import numpy as np
from .utility import Utility
from .sa import Sa
import os
import sys
import scipy.io as sio
import matplotlib.pyplot as plt
from matplotlib import colors
import mpl_toolkits.mplot3d.art3d as art3d


class AvgSa(Sa):
    large_err = 1.0e99

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, structure, gmm, correl_func, define_range, **kwargs):
        self.define_range = define_range
        self.range_multiplier = kwargs.get('range_multiplier', [1, 1])
        self.num_periods = kwargs.get('num_periods', 10)
        self.type_spacing = kwargs.get('type_spacing', 'log')
        super().__init__(structure, gmm, correl_func)

    ####################################################################################################################
    # Static methods
    ####################################################################################################################

    @staticmethod
    def evaluate_gmm_helper(med_sa_period_range, sig_ln_sa_period_range, moc):
        n_p = med_sa_period_range.shape[1]

        med_sa_avg = np.exp((1 / n_p) * np.sum(np.log(med_sa_period_range), axis=1))
        sig_ij = np.einsum('ij,il->jli', sig_ln_sa_period_range, sig_ln_sa_period_range)
        rho_ij_sig_ij = np.einsum('ijk,ij->ijk', sig_ij, moc)
        sig_ln_sa_avg = np.sqrt(((1 / n_p) ** 2) * np.sum(rho_ij_sig_ij, axis=(0, 1)))

        # return median and std
        return med_sa_avg, sig_ln_sa_avg

    ####################################################################################################################
    # Public Functionalities (instance methods)
    ####################################################################################################################

    def evaluate_period(self, for_which):
        structure = self.structure
        define_range = self.define_range
        range_multiplier = self.range_multiplier
        num_periods = self.num_periods
        type_spacing = self.type_spacing
        period_range = structure.get_period_range(for_which, define_range, range_multiplier, num_periods, type_spacing)
        return period_range

    # ------------------------------------------------------------------------------------------------------------------

    def setup_psa_spectra_gms(self, mrp, for_which, mean_req_exact, sig_req_exact, spectral_periods,
                              gm, sa_known, per_known, is_scaled):
        # Here zeros are ok as intialized values of indices because they are going to be filled up before they are used
        rec_per = np.zeros(len(spectral_periods)).astype(int)
        for i in range(len(spectral_periods)):
            rec_per[i] = np.argmin(np.abs(per_known - spectral_periods[i]))

        sa_known[sa_known == 0] = sys.float_info.epsilon
        sample_big = np.log(sa_known[:, rec_per])

        period_range = self.evaluate_period(for_which)
        target = self.get_inv_seismic_hazard(mrp, for_which)
        ind = np.in1d(spectral_periods, period_range)
        rec_avg = np.exp(np.sum(sample_big[:, ind], axis=1) / len(period_range))
        if is_scaled:
            scale_fac = target / rec_avg
        else:
            scale_fac = rec_avg / rec_avg

        return sample_big, scale_fac, mean_req_exact, sig_req_exact, gm

    # ------------------------------------------------------------------------------------------------------------------

    def evaluate_gmm(self, period_range):
        med_sa_period_range, sig_ln_sa_period_range = self._evaluate_gmm(period_range)
        moc = self.get_moc(period_range)
        med_sa_avg, sig_ln_sa_avg = AvgSa.evaluate_gmm_helper(med_sa_period_range, sig_ln_sa_period_range, moc)
        # return 1-d arrays
        return med_sa_avg, sig_ln_sa_avg

    # ------------------------------------------------------------------------------------------------------------------

    def _get_target_spectrum(self, mrp, for_which, med_sa_spectrum, sig_ln_sa_spectrum,
                             spectral_periods, uncondition):
        if uncondition:
            scenario_weights = self.get_seismic_hazard_deagg(mrp, for_which)
            mean_ln_sa_spectrum, sig_ln_sa_spectrum = Utility.get_exact_mixture_params(
                np.log(med_sa_spectrum),
                sig_ln_sa_spectrum,
                scenario_weights)  # (1, n_sp), (1, n_sp)
            med_sa_spectrum = np.exp(mean_ln_sa_spectrum)  # (1, n_sp)

        period_range = self.evaluate_period(for_which)

        n_s = med_sa_spectrum.shape[0]
        n_sp = len(spectral_periods)
        n_p = len(period_range)

        ind_period_range = np.in1d(spectral_periods, period_range)
        med_sa_period_range = med_sa_spectrum[:, ind_period_range]  # (n_s, n_p)
        sig_ln_sa_period_range = sig_ln_sa_spectrum[:, ind_period_range]  # (n_s, n_p)

        moc = self.get_moc(period_range)
        med_sa_avg, sig_ln_sa_avg = AvgSa.evaluate_gmm_helper(med_sa_period_range,
                                                              sig_ln_sa_period_range,
                                                              moc)  # (n_s,), (n_s,)

        rho_ij = np.zeros((n_sp, n_p))  # (n_sp, n_p)
        for j in range(n_p):
            rho_ij[:, j] = self.correl_func(spectral_periods, period_range[j])

        rho = (rho_ij @ sig_ln_sa_period_range.T).T  # ((n_sp, n_p) @ (n_p, n_s)).T = (n_s, n_sp)
        rho = rho / (n_p * sig_ln_sa_avg.reshape(n_s, 1))  # (n_s, n_sp) / (n_s, 1) = (n_s, n_sp)

        mean_req, cov_req = self.get_conditional_spectrum(mrp, for_which,
                                                          med_sa_spectrum,
                                                          sig_ln_sa_spectrum,
                                                          rho, self.get_moc(spectral_periods),
                                                          med_sa_avg, sig_ln_sa_avg)
        return mean_req, cov_req, spectral_periods

    # ------------------------------------------------------------------------------------------------------------------

    def plot_shc(self, for_which, **kwargs):
        save_dir_path = kwargs.get('save_dir_path', Utility.get_path(os.getcwd()))
        save_mat = kwargs.get('save_mat', False)
        lc = kwargs.get('lc', 'black')
        lw = kwargs.get('lw', 1.5)
        plot_all = kwargs.get('plot_all', False)

        fig, ax, export_mat_dict = super().plot_shc(for_which, **kwargs)
        structure = self.structure
        if plot_all:
            period_range = self.evaluate_period(for_which)
            shc = export_mat_dict['shc']
            shc_list = self.shc_at_periods(period_range, im_input=shc[:, 0])
            for itr in range(len(period_range)):
                shc_individual = shc_list[itr]
                ax.loglog(shc_individual[:, 0], shc_individual[:, 1], '-',
                          color=0.5 * (np.ones(3) - np.array(colors.to_rgb(lc))), alpha=0.5, linewidth=lw / 2)
                export_mat_dict[f'shc_individual_{itr + 1}'] = shc_individual
            if save_mat:
                name = structure.name
                for_which_str = '_'.join(for_which)
                file_name = f'plot_shc_{name}_{for_which_str}.mat'
                sio.savemat(Utility.get_path(save_dir_path, file_name), export_mat_dict)

        return fig, ax, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------

    def mark_period_on_axis(self, for_which, ax, **kwargs):
        patch_alpha = kwargs.get('period_patch_alpha', 0.25)
        patch_color = kwargs.get('period_patch_color', (0, 1, 0))
        period_range = self.evaluate_period(for_which)
        ylim = Utility.get_ylim(ax, (0.01, None))
        rect = plt.Rectangle(
            (period_range[0], ylim[0]), period_range[-1] - period_range[0], ylim[1] - ylim[0],
            color=patch_color, alpha=patch_alpha
        )
        ax.add_patch(rect)
        if '3D' in ax.__class__.__name__:
            art3d.pathpatch_2d_to_3d(rect, z=0)
        ax.set_ylim(ylim)



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 20 20:36:32 2019

@author: Angshuman Deb
"""

from .analysis import Analysis
from .utility import Utility
from .structure import Structure
import os
import numpy as np
from time import time
from pathos.multiprocessing import ProcessPool as Pool
from functools import partial
from benedict import benedict


class PSDemHA(Analysis):

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, edp_list, im, **kwargs):
        comp_env = kwargs.get('comp_env', 'local')
        self.edp_list = edp_list
        self.im = im
        super().__init__(self.__class__.__name__, im.structure, comp_env)

    ####################################################################################################################
    # Static methods
    ####################################################################################################################

    @staticmethod
    def run_parallel(setup_dir_path, compute_demand_hazard_integral_list,
                     haz_lev_list, n_gm_list, im, delta_input_list, min_max_scale_fac_list,
                     node_index, batch_index, analysis_index, run_case):

        start_time = time()

        for_which = list(run_case.astype(str))
        analysis_case = for_which[Structure.get_dir_level_index('Analysis_Case')]
        run_case_str = ' '.join(for_which)

        psdemha_results = list()
        for itr in range(len(compute_demand_hazard_integral_list)):
            temp = compute_demand_hazard_integral_list[itr](for_which, haz_lev_list, im, n_gm_list,
                                                            min_max_scale_fac=min_max_scale_fac_list[itr],
                                                            delta_input=delta_input_list[itr])
            psdemha_results.append(temp)

        status_file_name = f'status_analysis_index_{int(analysis_index)}' \
                           f'_node_{int(node_index)}_batch_{int(batch_index)}.txt'
        status_file_path = Utility.get_path(setup_dir_path, analysis_case, status_file_name)

        with open(status_file_path, 'w') as fid:
            fid.write(f"{run_case_str} : COMPLETED in {time() - start_time} seconds\n")

        return f'{run_case_str}', psdemha_results

    ####################################################################################################################
    # Public functionalities (instance methods)
    ####################################################################################################################

    def stage(self, analysis_case, design_num_list):

        results_dir_path = self.get_results_dir_path()

        for design_num in design_num_list:
            target_work_dir_path = Utility.get_path(results_dir_path, analysis_case, design_num)

            if not os.path.isdir(target_work_dir_path):
                os.makedirs(target_work_dir_path)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def setup(self, analysis_case, design_num_list, haz_lev_list, n_gm_list=None, **kwargs):
        run_time_limit = kwargs.get('run_time_limit', '01:00:00')
        allocation_name = kwargs.get('allocation_name', '')
        n_batch = kwargs.get('n_batch', 1)
        n_job = kwargs.get('n_job', 1)
        pool_size_factor = kwargs.get('pool_size_factor', 1)

        structure = self.structure

        self.stage(analysis_case, design_num_list)

        setup_dir_path = self.get_setup_dir_path()
        if not os.path.isdir(Utility.get_path(setup_dir_path, analysis_case)):
            os.makedirs(Utility.get_path(setup_dir_path, analysis_case))

        temp = list()
        for design_num in design_num_list:
            temp.append(
                Utility.get_multi_level_iterator_cases(
                    int(analysis_case),
                    int(design_num),
                    list(range(1, structure.get_random_model_param_vals([analysis_case, design_num]).shape[2] + 1)),
                    list(range(1, structure.get_random_model_param_vals([analysis_case, design_num]).shape[0] + 1)),
                )
            )
        analysis_list = np.vstack(temp)

        list_file_path = Utility.get_path(setup_dir_path, analysis_case, 'analysis_list.txt')
        Utility.save_array_as_text_file(list_file_path, analysis_list, fmt='%d',
                                        header=f'Total analyses to run = {analysis_list.shape[0]:d}')

        self.set_analysis_setup_info(
            analysis_case,
            {'design_num_list': design_num_list,
             'haz_lev_list': haz_lev_list,
             'n_gm_list': n_gm_list,
             }
        )

        if not self.comp_env == 'local':
            self.setup_non_local_run(analysis_case, n_batch, n_job, pool_size_factor, run_time_limit, allocation_name)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def run(self, analysis_case, pool_size, **kwargs):
        edp_list = self.edp_list
        n_batch = kwargs.get('n_batch', 1)
        n_job = kwargs.get('n_job', 1)
        node_index = kwargs.get('node_index', 1)
        batch_index = kwargs.get('batch_index', 1)
        delta_input_list = kwargs.get('delta_input_list', [np.array([])] * len(edp_list))
        min_max_scale_fac_list = kwargs.get('min_max_scale_fac_list', [[1, 1]] * len(edp_list))

        compute_demand_hazard_integral_list = [edp.compute_demand_hazard_integral for edp in edp_list]
        im = self.im

        setup_dir_path = self.get_setup_dir_path()
        results_dir_path = self.get_results_dir_path()
        analysis_list, analysis_range, n_sim = self.get_analysis_info(analysis_case, 'analysis_list.txt',
                                                                      n_batch, n_job, node_index, batch_index)

        analysis_setup_info = self.get_analysis_setup_info(analysis_case)
        design_num_list = list(np.unique(analysis_list[:, 1]).astype(str))
        haz_lev_list = analysis_setup_info['haz_lev_list']
        n_gm_list = analysis_setup_info['n_gm_list']

        to_run = partial(PSDemHA.run_parallel, setup_dir_path, compute_demand_hazard_integral_list,
                         haz_lev_list, n_gm_list, im, delta_input_list, min_max_scale_fac_list,
                         node_index, batch_index)

        to_pass = [[analysis_index for analysis_index in analysis_range],
                   [analysis_list[analysis_index - 1, :] for analysis_index in analysis_range], ]

        if pool_size > 1:
            my_pool = Pool(pool_size)
            psdemha_results = dict(my_pool.map(to_run, *to_pass))
            my_pool.close()
            my_pool.join()
            my_pool.clear()
        else:
            psdemha_results = dict(list(map(to_run, *to_pass)))

        edp_tags = [edp.tag for edp in edp_list]

        for design_num in design_num_list:
            target_work_dir_path = Utility.get_path(results_dir_path, analysis_case, design_num)
            psdemha_results_design_num = {key: psdemha_results[key] for key in psdemha_results.keys() if
                                          key.split()[Structure.get_dir_level_index('Design_Num')] == design_num}
            psdemha_results_edp = {
                edp_tag: {analysis_key: psdemha_results_design_num[analysis_key][edp_tags.index(edp_tag)]
                          for analysis_key in psdemha_results_design_num.keys()}
                for edp_tag in edp_tags
            }

            if n_batch == 1 and n_job == 1:
                Utility.pickle_dump_dict(Utility.get_path(target_work_dir_path, f'psdemha_results.pickle'),
                                         psdemha_results_edp)
            else:
                Utility.pickle_dump_dict(
                    Utility.get_path(
                        target_work_dir_path, f'psdemha_results_node_{node_index}_batch_{batch_index}.pickle'
                    ), psdemha_results_edp
                )

        # Write final_status_node_{}_batch_{}.txt
        contains_str = f'_node_{int(node_index)}_batch_{int(batch_index)}'
        Utility.combine_contents_of_files(Utility.get_path(setup_dir_path, analysis_case),
                                          [f'contains status_analysis_index_ {contains_str}'],
                                          f'final_status{contains_str}.txt',
                                          file_ext='.txt', delete_original=True)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def wrap_up(self, analysis_case):
        setup_dir_path = self.get_setup_dir_path()
        results_dir_path = self.get_results_dir_path()

        # Write final_status.txt
        Utility.combine_contents_of_files(Utility.get_path(setup_dir_path, analysis_case),
                                          ['startswith final_status_', 'startswith status_analysis_index_'],
                                          'final_status.txt',
                                          file_ext='.txt', delete_original=True)

        analysis_list, _, _ = self.get_analysis_info(analysis_case, 'analysis_list.txt', 1, 1, 1, 1)

        # Combine results from all nodes and batches
        design_num_list = list(np.unique(analysis_list[:, 1]).astype(str))

        for design_num in design_num_list:
            target_work_dir_path = Utility.get_path(results_dir_path, analysis_case, design_num)
            files = os.listdir(target_work_dir_path)
            files = [file for file in files if file.startswith('psdemha_results') and file.endswith(
                '.pickle') and 'node' in file and 'batch' in file]
            if len(files) > 0:
                list_of_dicts = [Utility.pickle_load_dict(Utility.get_path(target_work_dir_path, file)) for file in
                                 files]

                psdemha_results_file_path = Utility.get_path(target_work_dir_path, 'psdemha_results.pickle')
                if os.path.isfile(psdemha_results_file_path):
                    psdemha_results = benedict(Utility.pickle_load_dict(psdemha_results_file_path))
                else:
                    psdemha_results = benedict()

                for d in list_of_dicts:
                    psdemha_results.merge(d)

                Utility.pickle_dump_dict(psdemha_results_file_path, psdemha_results)

                for file in files:
                    os.remove(Utility.get_path(target_work_dir_path, file))

        return

    # ------------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 20 20:36:32 2019

@author: Angshuman Deb
"""

from .analysis import Analysis
from .utility import Utility
from .structure import Structure
import os
import numpy as np
from time import time
from pathos.multiprocessing import ProcessPool as Pool
from functools import partial
import shutil
from copy import deepcopy
# import platform


class NLTHA(Analysis):

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, edp_list, im, **kwargs):
        comp_env = kwargs.get('comp_env', 'local')
        self.edp_list = edp_list
        self.im = im
        super().__init__(self.__class__.__name__, edp_list[0].structure, comp_env)

    ####################################################################################################################
    # Static methods
    ####################################################################################################################

    @staticmethod
    def stage_parallel(results_dir_path, model_fetch_dir_path, write_run_nltha,
                       generate_recorder_list, rec_gen_file_open_mode_list,
                       gen_rec_list, get_gm_rec_info, gm_database_dir_path, ai_end,
                       run_case, model_fetch_case, gm_fetch_case):

        for_which = list(run_case.astype(str))
        model_fetch_for_which = list(model_fetch_case.astype(str))
        gm_fetch_for_which = list(gm_fetch_case.astype(str))

        target_work_dir_path = Utility.get_path(results_dir_path, *for_which)
        work_dir_path_prelim_analysis = Utility.get_path(model_fetch_dir_path, *model_fetch_for_which)

        # make target work dir if it doesn't exist
        if not os.path.isdir(target_work_dir_path):
            os.makedirs(target_work_dir_path)

        status_file_path = Utility.get_path(target_work_dir_path, 'NLTHA_STATUS.txt')
        if os.path.isfile(status_file_path):
            os.remove(status_file_path)

        # generate edp recorders
        for i_edp in range(len(generate_recorder_list)):
            generate_recorder_list[i_edp](for_which, rec_gen_file_open_mode=rec_gen_file_open_mode_list[i_edp],
                                          gen_rec=gen_rec_list[i_edp])

        # write run_nltha
        write_run_nltha(target_work_dir_path, work_dir_path_prelim_analysis)

        # fetch ground motions
        if gm_database_dir_path is not None:
            gm_file_names, scale_fac = get_gm_rec_info([*gm_fetch_for_which,
                                                        for_which[
                                                            Structure.get_dir_level_index('Ground_Motion_Rec_Num')]])
            Utility.fetch_selected_ground_motion_records(
                gm_database_dir_path, gm_file_names, scale_fac, target_work_dir_path, ai_end)

        run_case_str = ' '.join(for_which)
        print(f'Staging analysis case {run_case_str} complete!')

        return

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def run_parallel(results_dir_path, setup_dir_path, run_nltha, comp_env, node_index, batch_index,
                     analysis_index, run_case):
        start_time = time()

        for_which = list(run_case.astype(str))
        analysis_case = for_which[Structure.get_dir_level_index('Analysis_Case')]
        run_case_str = ' '.join(for_which)

        target_work_dir_path = Utility.get_path(results_dir_path, *for_which)

        run_nltha(target_work_dir_path, comp_env)

        status = 'FAIL'
        nltha_status_file_path = Utility.get_path(target_work_dir_path, 'NLTHA_STATUS.txt')
        if os.path.isfile(nltha_status_file_path):
            with open(nltha_status_file_path, 'r') as fid:
                status = fid.readline().strip()

        status_file_name = f'status_analysis_index_{int(analysis_index)}' \
                           f'_node_{int(node_index)}_batch_{int(batch_index)}.txt'
        status_file_path = Utility.get_path(setup_dir_path, analysis_case, status_file_name)

        with open(status_file_path, 'w') as fid:
            if status == 'SUCCESS':
                fid.write(f"{run_case_str} : CONVERGED and COMPLETED in {time() - start_time} seconds\n")
            else:
                fid.write(f"{run_case_str} : FAILED in {time() - start_time} seconds\n")

        return

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def run_parallel_prelim_analysis(results_dir_path, comp_env, run_prelim_analysis, run_case):
        for_which = list(run_case.astype(str))
        target_work_dir_path = Utility.get_path(results_dir_path, *for_which)
        run_prelim_analysis(target_work_dir_path, comp_env)
        return

    ####################################################################################################################
    # Public functionalities (instance methods)
    ####################################################################################################################

    def stage(self, pool_size, analysis_list, model_fetch_list, gm_fetch_list, gm_database_dir_path, **kwargs):

        ai_end = kwargs.get('ai_end', 0.7)

        edp_list = self.edp_list
        results_dir_path = self.get_results_dir_path()
        model_fetch_dir_path = results_dir_path.replace(self.analysis_type, 'Prelim_Analysis')

        for edp in edp_list:
            edp.write_evaluate_edp_and_helpers()

        generate_recorder_list = list()
        rec_gen_file_open_mode_list = ['w'] + ['a+'] * (len(edp_list) - 1)

        gen_rec_list = np.array([True] * len(edp_list))
        ind_shared = np.array([edp.recorder_file_storage == 'shared' for edp in edp_list])
        ind_shared[0] = False
        gen_rec_list[ind_shared] = False
        gen_rec_list = list(gen_rec_list)

        for edp in edp_list:
            generate_recorder_list.append(edp.generate_recorder)

        get_gm_rec_info = self.im.get_gm_rec_info

        n_sim = analysis_list.shape[0]

        to_run = partial(NLTHA.stage_parallel, results_dir_path, model_fetch_dir_path,
                         self.structure.structural_analysis_platform.write_run_nltha,
                         generate_recorder_list, rec_gen_file_open_mode_list,
                         gen_rec_list, get_gm_rec_info, gm_database_dir_path, ai_end)

        to_pass = [[analysis_list[itr, :] for itr in range(n_sim)],
                   [model_fetch_list[itr, :] for itr in range(n_sim)],
                   [gm_fetch_list[itr, :] for itr in range(n_sim)], ]

        if pool_size > 1:
            my_pool = Pool(pool_size)
            my_pool.map(to_run, *to_pass)
            my_pool.close()
            my_pool.join()
            my_pool.clear()
        else:
            list(map(to_run, *to_pass))

        return

    # ------------------------------------------------------------------------------------------------------------------

    def setup(self, analysis_case, pool_size, design_num_list, haz_lev_list, gm_database_dir_path,
              **kwargs):
        run_time_limit = kwargs.get('run_time_limit', '01:00:00')
        allocation_name = kwargs.get('allocation_name', '')
        n_batch = kwargs.get('n_batch', 1)
        n_job = kwargs.get('n_job', 1)
        pool_size_factor = kwargs.get('pool_size_factor', 1)
        stage = kwargs.get('stage', True)
        ai_end = kwargs.get('ai_end', 0.7)

        structure = self.structure
        im = self.im

        setup_dir_path = self.get_setup_dir_path()
        if not os.path.isdir(Utility.get_path(setup_dir_path, analysis_case)):
            os.makedirs(Utility.get_path(setup_dir_path, analysis_case))

        temp = list()
        for design_num in design_num_list:
            for haz_lev in haz_lev_list:
                _, _, _, n_gm = im.get_gms_results([analysis_case, design_num, '1', '1', haz_lev])
                temp.append(
                    Utility.get_multi_level_iterator_cases(
                        int(analysis_case),
                        int(design_num),
                        list(range(1, structure.get_random_model_param_vals([analysis_case, design_num]).shape[2] + 1)),
                        list(range(1, structure.get_random_model_param_vals([analysis_case, design_num]).shape[0] + 1)),
                        int(haz_lev),
                        list(range(1, n_gm + 1))
                    )
                )
        analysis_list = np.vstack(temp)
        model_fetch_list = analysis_list[:, :Structure.get_dir_level_index('Model_Realization_Num') + 1]
        gm_fetch_list = analysis_list[:, :Structure.get_dir_level_index('Hazard_Level_Num') + 1]

        list_file_path = Utility.get_path(setup_dir_path, analysis_case, 'analysis_list.txt')
        list_file_path_orig = Utility.get_path(setup_dir_path, analysis_case, 'analysis_list_original.txt')
        Utility.save_array_as_text_file(list_file_path, analysis_list, fmt='%d',
                                        header=f'Total analyses to run = {analysis_list.shape[0]:d}')
        Utility.save_array_as_text_file(list_file_path_orig, analysis_list, fmt='%d',
                                        header=f'Total analyses to run = {analysis_list.shape[0]:d}')

        if stage:
            self.stage(pool_size, analysis_list.astype(int), model_fetch_list.astype(int),
                       gm_fetch_list.astype(int), gm_database_dir_path, ai_end=ai_end)

        self.set_analysis_setup_info(
            analysis_case,
            {'design_num_list': design_num_list,
             'haz_lev_list': haz_lev_list,
             'allocation_name': allocation_name
             }
        )

        if not self.comp_env == 'local':
            self.setup_non_local_run(analysis_case, n_batch, n_job, pool_size_factor, run_time_limit, allocation_name)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def run(self, analysis_case, pool_size, **kwargs):
        n_batch = kwargs.get('n_batch', 1)
        n_job = kwargs.get('n_job', 1)
        node_index = kwargs.get('node_index', 1)
        batch_index = kwargs.get('batch_index', 1)
        design_nbrs = kwargs.get('design_nbrs', list())
        model_opts = kwargs.get('model_opts', list())
        model_realizations = kwargs.get('model_realizations', list())
        haz_levs = kwargs.get('haz_levs', list())
        gm_nbrs = kwargs.get('gm_nbrs', list())
        restage = kwargs.get('restage', False)

        include_list = [design_nbrs, model_opts, model_realizations, haz_levs, gm_nbrs]

        setup_dir_path = self.get_setup_dir_path()
        results_dir_path = self.get_results_dir_path()
        analysis_list, analysis_range, n_sim = self.get_analysis_info(analysis_case, 'analysis_list.txt', n_batch,
                                                                      n_job, node_index, batch_index)

        structure = self.structure
        run_nltha = structure.structural_analysis_platform.run_nltha
        analysis_ind = np.ones((analysis_list.shape[0],), dtype=bool)

        dir_levels = Structure.get_all_dir_levels()[Structure.get_dir_level_index('Analysis_Case') + 1:]

        for itr in range(len(dir_levels)):
            if len(include_list[itr]) == 0:
                include_list[itr] = np.unique(analysis_list[:, Structure.get_dir_level_index(dir_levels[itr])])
            else:
                include_list[itr] = np.array(include_list[itr]).astype(int)

        for itr in range(len(dir_levels)):
            if not np.array_equal(np.sort(include_list[itr]),
                                  np.unique(analysis_list[:, Structure.get_dir_level_index(dir_levels[itr])])):
                temp = ~np.ones((analysis_list.shape[0],), dtype=bool)
                for incl_item in include_list[itr]:
                    temp |= analysis_list[:, itr + 1] == incl_item
                analysis_ind &= temp

        if not np.all(analysis_ind):
            analysis_list = analysis_list[analysis_ind, :]
            analysis_range = np.arange(1, len(analysis_list) + 1)
            restage = True

        if restage:
            self.restage(analysis_case, pool_size, None, analysis_list=analysis_list)

        to_run = partial(NLTHA.run_parallel, results_dir_path, setup_dir_path, run_nltha,
                         self.comp_env, node_index, batch_index)

        to_pass = [[analysis_index for analysis_index in analysis_range],
                   [analysis_list[analysis_index - 1, :] for analysis_index in analysis_range], ]

        if pool_size > 1:
            my_pool = Pool(pool_size)
            my_pool.map(to_run, *to_pass)
            my_pool.close()
            my_pool.join()
            my_pool.clear()
        else:
            list(map(to_run, *to_pass))

        # Write final_status_node_{}_batch_{}.txt
        contains_str = f'_node_{int(node_index)}_batch_{int(batch_index)}'
        Utility.combine_contents_of_files(Utility.get_path(setup_dir_path, analysis_case),
                                          [f'contains status_analysis_index_ {contains_str}'],
                                          f'final_status{contains_str}.txt',
                                          file_ext='.txt', delete_original=True)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def wrap_up(self, analysis_case):
        setup_dir_path = self.get_setup_dir_path()

        # Write final_status.txt
        Utility.combine_contents_of_files(Utility.get_path(setup_dir_path, analysis_case),
                                          ['startswith final_status_', 'startswith status_analysis_index_'],
                                          'final_status.txt',
                                          file_ext='.txt', delete_original=True)

        analysis_ran, analysis_failed, analysis_skipped = self.get_post_analysis_lists(analysis_case)

        shutil.copy(Utility.get_path(setup_dir_path, analysis_case, 'analysis_list.txt'),
                    Utility.get_path(setup_dir_path, analysis_case, 'analysis_list_last.txt'))

        analysis_list_rerun = np.vstack([analysis_skipped, analysis_failed])
        if analysis_list_rerun.size != 0:
            analysis_list_rerun = np.unique(analysis_list_rerun, axis=0)
        self.setup_rerun(analysis_case, analysis_list_rerun)
        return

    # ------------------------------------------------------------------------------------------------------------------

    def setup_rerun(self, analysis_case, analysis_list_rerun, **kwargs):
        stage = kwargs.get('stage', False)
        pool_size = kwargs.get('pool_size', 1)
        gm_database_dir_path = kwargs.get('gm_database_dir_path', None)

        setup_dir_path = self.get_setup_dir_path()
        new_list_file_path = Utility.get_path(setup_dir_path, analysis_case, 'analysis_list.txt')
        Utility.save_array_as_text_file(new_list_file_path, analysis_list_rerun, fmt='%d',
                                        header=f'Total analyses to run = {analysis_list_rerun.shape[0]:d}')

        num_rerun = analysis_list_rerun.shape[0]
        if num_rerun > 0:
            rerun_stampede = input(f'{num_rerun} NLTHA did not converge! Re-run on Stampede2 [y/n] ? : ')
            if rerun_stampede == 'y' or rerun_stampede == 'Y':
                comp_env = self.comp_env
                analysis_setup_info = self.get_analysis_setup_info(analysis_case)
                allocation_name = analysis_setup_info['allocation_name']
                comp_env_rerun = input(f'Enter comp_env to re-run (default = \'{comp_env}\'): ').strip(
                    '"\'') or comp_env
                self.comp_env = comp_env_rerun
                n_batch = int(input('Enter number of batches to re-run (default = 1): ').strip('"\'') or '1')
                n_job = int(input('Enter number of nodes per batch to re-run (default = 1): ').strip('"\'') or '1')
                pool_size_factor = input('Enter pool_size_factor to re-run (default = 1): ').strip('"\'') or '1'
                if '/' in pool_size_factor:
                    frac = pool_size_factor.split('/')
                    numerator = float(frac[0])
                    denominator = float(frac[1])
                    pool_size_factor = numerator / denominator
                else:
                    pool_size_factor = float(pool_size_factor)
                run_time_limit = input('Enter run_time_limit to re-run (default = \'01:00:00\'): ').strip(
                    '"\'') or '01:00:00'
                allocation_name = input(f'Enter allocation_name to re-run (default = \'{allocation_name}\'): ').strip(
                    '"\'') or allocation_name
                self.setup_non_local_run(analysis_case, n_batch, n_job, pool_size_factor, run_time_limit,
                                         allocation_name)
                if stage:
                    self.restage(analysis_case, pool_size, gm_database_dir_path)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def recheck_status(self, analysis_case, **kwargs):
        print_status_file = kwargs.get('print_status_file', True)

        setup_dir_path = self.get_setup_dir_path()
        results_dir_path = self.get_results_dir_path()

        analysis_list, _, n_sim = self.get_analysis_info(analysis_case, 'analysis_list_original.txt', 1, 1, 1, 1)

        conv_count = 0

        write_lines = list()

        for i_sim in range(n_sim):
            run_case = analysis_list[i_sim, :]
            for_which = list(run_case.astype(str))
            run_case_str = ' '.join(for_which)
            target_work_dir_path = Utility.get_path(results_dir_path, *for_which)
            status = 'FAIL'
            nltha_status_file_path = Utility.get_path(target_work_dir_path, 'NLTHA_STATUS.txt')
            if os.path.isfile(nltha_status_file_path):
                with open(nltha_status_file_path, 'r') as nltha_status_fid:
                    status = nltha_status_fid.readline().strip()

            if status == 'SUCCESS':
                conv_count += 1
                if print_status_file:
                    write_lines.append(f"{run_case_str} : CONVERGED and COMPLETED in unknown seconds\n")
            else:
                if print_status_file:
                    write_lines.append(f"{run_case_str} : FAILED in unknown seconds\n")

        if print_status_file:
            status_file_path = Utility.get_path(setup_dir_path, analysis_case, 'final_status.txt')
            fid = open(status_file_path, 'w')
            fid.writelines(write_lines)
            fid.close()

        return conv_count, n_sim

    # ------------------------------------------------------------------------------------------------------------------

    def restage(self, analysis_case, pool_size, gm_database_dir_path, **kwargs):
        analysis_list = kwargs.get('analysis_list', None)
        ai_end = kwargs.get('ai_end', 0.7)

        structure = self.structure
        results_dir_path = self.get_results_dir_path().replace(self.analysis_type, 'Prelim_Analysis')

        if analysis_list is None:
            analysis_list, _, _ = self.get_analysis_info(analysis_case, 'analysis_list.txt', 1, 1, 1, 1)

        if analysis_list.size == 0:
            return

        model_fetch_list = analysis_list[:, :Structure.get_dir_level_index('Model_Realization_Num') + 1]
        gm_fetch_list = analysis_list[:, :Structure.get_dir_level_index('Hazard_Level_Num') + 1]

        prelim_analysis_list = np.unique(model_fetch_list, axis=0)

        to_run = partial(NLTHA.run_parallel_prelim_analysis, results_dir_path, self.comp_env,
                         structure.structural_analysis_platform.run_prelim_analysis)
        to_pass = [[prelim_analysis_list[analysis_index - 1, :]
                    for analysis_index in range(1, prelim_analysis_list.shape[0] + 1)], ]

        if pool_size > 1:
            my_pool = Pool(pool_size)
            my_pool.map(to_run, *to_pass)
            my_pool.close()
            my_pool.join()
            my_pool.clear()
        else:
            list(map(to_run, *to_pass))

        for analysis_index in range(1, prelim_analysis_list.shape[0] + 1):
            run_case = prelim_analysis_list[analysis_index - 1, :]
            for_which = list(run_case.astype(str))
            structure.write_damping_file(for_which)

        self.stage(pool_size, analysis_list.astype(int), model_fetch_list.astype(int),
                   gm_fetch_list.astype(int), gm_database_dir_path, ai_end=ai_end)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def find_anomalous_run_cases(self, analysis_case, threshold_list, **kwargs):
        save_analysis_list = kwargs.get('save_analysis_list', False)
        edp_list = self.edp_list
        analysis_list, analysis_range, _ = self.get_analysis_info(analysis_case,
                                                                  'analysis_list_original.txt', 1, 1, 1, 1)
        design_num_list = list(np.unique(analysis_list[:, 1]).astype(str))
        run_cases = np.array([]).reshape(0, len(Structure.get_all_dir_levels()))
        for itr in range(len(edp_list)):
            edp = edp_list[itr]
            for design_num in design_num_list:
                run_cases = np.vstack([run_cases,
                                       edp.find_anomalous_run_cases([analysis_case, design_num],
                                                                    threshold_list[itr])])

        if run_cases.shape[0] == 0:
            run_cases = run_cases.reshape(0, len(Structure.get_all_dir_levels()))
        else:
            run_cases = np.unique(run_cases, axis=0)

        if save_analysis_list:
            save_path = Utility.get_path(self.get_setup_dir_path(), analysis_case, 'analysis_list_anomalous.txt')
            Utility.save_array_as_text_file(save_path, run_cases, fmt='%d',
                                            header=f'Total analyses to run = {run_cases.shape[0]:d}')

        return run_cases

    # ------------------------------------------------------------------------------------------------------------------

    def extract_single_run(self, for_which, num_modes, gm_database_dir_path, **kwargs):

        recorder_info = kwargs.get('recorder_info', {})
        addnl_rec_file_open_mode = kwargs.get('addnl_rec_file_open_mode', 'w')
        ai_end = kwargs.get('ai_end', 0.7)

        structure = self.structure
        name = structure.name
        model_files_path = structure.model_files_path

        for_which1 = Structure.get_modified_for_which(for_which, 'Model_Realization_Num')
        for_which2 = Structure.get_modified_for_which(for_which, 'Hazard_Level_Num')

        base_dir_path = os.path.dirname(model_files_path)
        analysis_platform = structure.structural_analysis_platform.__class__.__name__
        extract_dest_path = Utility.get_path(base_dir_path, f'{name}_extract_{analysis_platform}')
        if not os.path.isdir(extract_dest_path):
            os.makedirs(extract_dest_path)

        # PrelimAnalysis extract
        NLTHA.run_parallel_prelim_analysis(
            self.get_results_dir_path().replace(self.analysis_type, 'Prelim_Analysis'),
            'local',
            structure.structural_analysis_platform.run_prelim_analysis,
            np.array(for_which1).astype(int))
        structure.extract_prelim_analysis(for_which, extract_dest_path, num_modes)

        # NLTHA extract
        self.stage(1,
                   np.array([for_which]).astype(int),
                   np.array([for_which1]).astype(int),
                   np.array([for_which2]).astype(int),
                   None,
                   ai_end=ai_end)

        # GM files fetch
        gm_file_names, scale_fac = self.im.get_gm_rec_info(for_which)
        Utility.fetch_selected_ground_motion_records(gm_database_dir_path, gm_file_names, scale_fac, extract_dest_path, ai_end)
        structure.extract_nltha(for_which, extract_dest_path)

        # ModelFiles extract
        stuff_in_model_files_path = os.listdir(model_files_path)
        [shutil.copy(Utility.get_path(model_files_path, file), extract_dest_path)
         for file in stuff_in_model_files_path if
         os.path.isfile(Utility.get_path(model_files_path, file)) and file.endswith(structure.structural_analysis_platform.file_ext)]

        # AdditionalRecorders
        structure.structural_analysis_platform.write_additional_recorders(recorder_info, extract_dest_path, addnl_rec_file_open_mode)

        # Leave a bat file for easy execution
        structure.structural_analysis_platform.bat_file(extract_dest_path)
        return

    # ------------------------------------------------------------------------------------------------------------------

    def insert_extracted_run(self, for_which):
        structure = self.structure
        name = structure.name
        model_files_path = structure.model_files_path

        base_dir_path = os.path.dirname(model_files_path)
        analysis_platform = structure.structural_analysis_platform.__class__.__name__
        src_path = Utility.get_path(base_dir_path, f'{name}_extract_{analysis_platform}')

        results_dir_path = self.get_results_dir_path()
        work_dir_path = Utility.get_path(results_dir_path, *for_which)

        edp_tags = [edp.tag for edp in self.edp_list]
        for tag in edp_tags:
            stuff_in_edp_results = os.listdir(Utility.get_path(src_path, f'EDP_{tag}_Results'))
            dest_path = Utility.get_path(work_dir_path, f'EDP_{tag}_Results')
            if not os.path.isdir(dest_path):
                os.makedirs(dest_path)
            [shutil.copy(Utility.get_path(src_path, f'EDP_{tag}_Results', file),
                         dest_path)
             for file in stuff_in_edp_results if
             os.path.isfile(Utility.get_path(src_path, f'EDP_{tag}_Results', file))]

        with open(Utility.get_path(work_dir_path, 'NLTHA_STATUS.txt'), 'w') as fid:
            fid.write('SUCCESS\n')

        analysis_case = for_which[Structure.get_dir_level_index('Analysis_Case')]
        final_status_file_path = Utility.get_path(self.get_setup_dir_path(), analysis_case, 'final_status.txt')
        if os.path.isfile(final_status_file_path):
            with open(final_status_file_path, 'r') as fid:
                lines = fid.readlines()
            for_which_str = ' '.join(for_which)
            inds = [i for i in range(len(lines)) if
                    lines[i].startswith(f"{for_which_str} : FAILED in ")
                    ]
            for ind in inds:
                lines[ind] = f"{for_which_str} : CONVERGED and COMPLETED in unknown seconds\n"
            if len(inds) == 0:
                lines.append(f"{for_which_str} : CONVERGED and COMPLETED in unknown seconds\n")
            with open(final_status_file_path, 'w') as fid:
                fid.writelines(lines)

        [edp.collect(analysis_case, 1, analysis_list=np.array([[int(item) for item in for_which]]))
         for edp in self.edp_list]

        return

    # ------------------------------------------------------------------------------------------------------------------

    def stampede_upload_pre(self, analysis_case, design_num_list, haz_lev_list, username, allocation_name, n_batch, n_job, upload_to, **kwargs):
        local_bash_path = kwargs.get('local_bash_path', '')
        comp_env = kwargs.get('comp_env', 'local')
        new_file = kwargs.get('new_file', False)
        keep_file = kwargs.get('keep_file', False)
        nltha_include = kwargs.get('nltha_include', False)
        extract = kwargs.get('extract', True)
        stampede_comp_env = kwargs.get('stampede_comp_env', 'stampede_knl')
        nltha_setup_pool_size = kwargs.get('nltha_setup_pool_size', 50)
        gm_database_dir_path = kwargs.get('gm_database_dir_path', Utility.get_path(os.path.dirname(upload_to), 'NGA_PEER_EQ_DB'))
        nltha_run_pool_size_factor = kwargs.get('nltha_run_pool_size_factor', 0.85)

        structure = self.structure
        include_list = [f"GMS_Results/{analysis_case}/{{{','.join(design_num_list)}}}",
                        f"GMS_Setup/{analysis_case}",
                        f"Prelim_Analysis_Results/{analysis_case}/{{{','.join(design_num_list)}}}",
                        f"Prelim_Analysis_Setup/{analysis_case}",
                        f"PSHA_Results/{analysis_case}/{{{','.join(design_num_list)}}}",
                        f"PSHA_Setup/{analysis_case}",
                        f"Random_Model_Param_Vals/{analysis_case}/{{{','.join(design_num_list)}}}"]
        if nltha_include:
            include_list.append(f"NLTHA_EDP/{analysis_case}/{{{','.join(design_num_list)}}}")
            include_list.append(f"NLTHA_Setup/{analysis_case}")

        if len(design_num_list) == 1:
            include_list = [item.replace('{', '').replace('}', '') for item in include_list]
        temp_file_name = structure.tar_work_dir(local_bash_path=local_bash_path, comp_env=comp_env, new_file=True,
                                                flags=['cvf'],
                                                include_list=include_list)

        shutil.move(Utility.get_path(structure.model_work_dir_path, temp_file_name),
                    structure.model_files_path)

        exec_commands = [f'tar -uvf {temp_file_name} * --exclude={structure.name}_Work_Dir*.tar']

        Utility.shell_exec('APPENDTAR', structure.model_files_path, exec_commands,
                           comp_env=comp_env, local_bash_path=local_bash_path)

        stampede_files_dir = Utility.get_path(os.getcwd(), 'Stampede_Files')
        os.makedirs(stampede_files_dir, exist_ok=True)
        objects_dict = {
            'analysis_case': analysis_case,
            'nltha_setup_pool_size': nltha_setup_pool_size,
            'gm_database_dir_path': gm_database_dir_path,
            'design_num_list': design_num_list,
            'haz_lev_list': haz_lev_list,
            'allocation_name': allocation_name,
            'n_batch': n_batch,
            'n_job': n_job,
            'nltha_run_pool_size_factor': nltha_run_pool_size_factor,
            'nltha': self.get_nltha_for_stampede_upload(upload_to, stampede_comp_env),
        }
        Utility.pickle_dump_dict(Utility.get_path(stampede_files_dir, 'objects.pickle'), objects_dict)

        with open(Utility.get_path(stampede_files_dir, 'setup_nltha.py'), 'w') as fid:
            fid.write('from pypbee.utility import Utility\n')
            fid.write('\n')
            fid.write(f"objects = Utility.pickle_load_dict(\'{Utility.get_path(upload_to, 'objects.pickle')}\')\n")
            fid.write('analysis_case = objects[\'analysis_case\']\n')
            fid.write('nltha_setup_pool_size = objects[\'nltha_setup_pool_size\']\n')
            fid.write('gm_database_dir_path = objects[\'gm_database_dir_path\']\n')
            fid.write('design_num_list = objects[\'design_num_list\']\n')
            fid.write('haz_lev_list = objects[\'haz_lev_list\']\n')
            fid.write('allocation_name = objects[\'allocation_name\']\n')
            fid.write('n_batch = objects[\'n_batch\']\n')
            fid.write('n_job = objects[\'n_job\']\n')
            fid.write('nltha_run_pool_size_factor = objects[\'nltha_run_pool_size_factor\']\n')
            fid.write('nltha = objects[\'nltha\']\n')
            fid.write('kwargs = {\n')
            fid.write('    \'run_time_limit\': \'01:00:00\',\n')
            fid.write('    \'allocation_name\': allocation_name,\n')
            fid.write('    \'n_batch\': n_batch,\n')
            fid.write('    \'n_job\': n_job,\n')
            fid.write('    \'pool_size_factor\': nltha_run_pool_size_factor,\n')
            fid.write('}\n')
            fid.write('nltha.setup(\n')
            fid.write('    analysis_case,\n')
            fid.write('    nltha_setup_pool_size,\n')
            fid.write('    design_num_list,\n')
            fid.write('    haz_lev_list,\n')
            fid.write('    gm_database_dir_path,\n')
            fid.write('    **kwargs\n')
            fid.write(')\n')

        with open(Utility.get_path(stampede_files_dir, 'run_on_stampede.py'), 'w') as fid:
            fid.write('from pypbee.utility import Utility\n')
            fid.write('\n')
            fid.write(f"objects = Utility.pickle_load_dict(\'{Utility.get_path(upload_to, 'objects.pickle')}\')\n")
            fid.write('nltha = objects[\'nltha\']\n')
            fid.write('\n')
            fid.write('\n')
            fid.write('def run_on_stampede_nltha(analysis_case, n_batch, n_job, node_index, batch_index, pool_size):\n')
            fid.write('    nltha.run(\n')
            fid.write('        analysis_case,\n')
            fid.write('        pool_size,\n')
            fid.write('        n_batch=n_batch,\n')
            fid.write('        n_job=n_job,\n')
            fid.write('        node_index=node_index,\n')
            fid.write('        batch_index=batch_index\n')
            fid.write('    )\n')

        shutil.move(Utility.get_path(structure.model_files_path, temp_file_name), stampede_files_dir)

        Utility.shell_exec('APPENDTAR', stampede_files_dir, exec_commands, comp_env=comp_env, local_bash_path=local_bash_path)

        shutil.move(Utility.get_path(stampede_files_dir, temp_file_name), Utility.get_path(os.getcwd()))

        shutil.rmtree(stampede_files_dir)

        Utility.shell_exec('COMPRESS', '',
                           ['echo "compressing ..."', f'gzip {temp_file_name}'],
                           comp_env=comp_env, local_bash_path=local_bash_path)

        if new_file:
            new_file_name = f'{structure.name}_nltha_{analysis_case}_stampede_upload_pre_{Utility.get_curr_time_stamp()}.tar'
        else:
            new_file_name = f'{structure.name}_nltha_{analysis_case}_stampede_upload_pre.tar'

        os.replace(temp_file_name + '.gz', new_file_name)

        src = new_file_name
        server = f"{username}@stampede2.tacc.utexas.edu"

        Utility.shell_exec('UPLOAD', '',
                           [f'scp {src} {server}:{upload_to}'],
                           comp_env=comp_env, local_bash_path=local_bash_path)

        if extract:
            Utility.shell_exec('EXTRACT', '',
                               [f"ssh -T {server} << EOF",
                                f'cd {upload_to}',
                                f'tar -xzvf {new_file_name}',
                                f'rm -rf {new_file_name}\n' if not keep_file else '',
                                'EOF'],
                               comp_env=comp_env, local_bash_path=local_bash_path)

        if not keep_file:
            os.remove(new_file_name)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def get_nltha_for_stampede_upload(self, upload_to, comp_env):
        new_structure = deepcopy(self.structure)
        new_structure.model_files_path = upload_to
        new_structure.model_work_dir_path = upload_to
        new_im = deepcopy(self.im)
        new_im.structure = new_structure
        new_edps = deepcopy(self.edp_list)
        for edp in new_edps:
            edp.structure = new_structure
            edp.haz_req = {}
        return NLTHA(new_edps, new_im, comp_env=comp_env)

    def get_post_analysis_lists(self, analysis_case):
        setup_dir_path = self.get_setup_dir_path()
        final_status_file_path = Utility.get_path(setup_dir_path, analysis_case, 'final_status.txt')

        with open(final_status_file_path, 'r') as fid:
            analysis_failed = np.array(
                [np.array(line.split(':')[0].strip().split()).astype(int)
                 for line in fid.readlines() if not line.isspace() and 'FAILED' in line])

        if analysis_failed.size == 0:
            analysis_failed = analysis_failed.reshape(0, len(Structure.get_all_dir_levels()))

        analysis_ran = Utility.read_numpy_array_from_txt_file(final_status_file_path, comments=':').astype(int)

        analysis_list, _, n_sim = self.get_analysis_info(analysis_case, 'analysis_list.txt', 1, 1, 1, 1)

        if n_sim > 0:
            analysis_skipped = analysis_list[Utility.in_nd(analysis_list, analysis_ran) == 0, :]
        else:
            analysis_skipped = np.array([]).reshape(0, len(Structure.get_all_dir_levels()))

        return analysis_ran, analysis_failed, analysis_skipped

    # ------------------------------------------------------------------------------------------------------------------

    def collect_edps(self, analysis_case, pool_size, analysis_list='analysis_list_original.txt'):
        edp_list = self.edp_list
        [edp.collect(analysis_case, pool_size, analysis_list=analysis_list) for edp in edp_list]

    # ------------------------------------------------------------------------------------------------------------------

    def stampede_download_post(self, analysis_case, pool_size, **kwargs):
        local_bash_path = kwargs.get('local_bash_path', '')
        comp_env = kwargs.get('comp_env', 'local')
        new_file = kwargs.get('new_file', False)
        analysis_list = kwargs.get('analysis_list', 'analysis_list_original.txt')

        self.collect_edps(analysis_case, pool_size, analysis_list=analysis_list)
        if isinstance(analysis_list, str):
            analysis_list, _, _ = self.get_analysis_info(analysis_case, analysis_list, 1, 1, 1, 1)
        design_num_list = list(np.unique(analysis_list[:, 1]).astype(str))

        structure = self.structure
        include_list = [f"GMS_Results/{analysis_case}/{{{','.join(design_num_list)}}}",
                        f"GMS_Setup/{analysis_case}",
                        f"Prelim_Analysis_Results/{analysis_case}/{{{','.join(design_num_list)}}}",
                        f"Prelim_Analysis_Setup/{analysis_case}",
                        f"PSHA_Results/{analysis_case}/{{{','.join(design_num_list)}}}",
                        f"PSHA_Setup/{analysis_case}",
                        f"Random_Model_Param_Vals/{analysis_case}/{{{','.join(design_num_list)}}}",
                        f"NLTHA_EDP/{analysis_case}/{{{','.join(design_num_list)}}}",
                        f"NLTHA_Setup/{analysis_case}"]
        if len(design_num_list) == 1:
            include_list = [item.replace('{', '').replace('}', '') for item in include_list]
        temp_file_name = structure.tar_work_dir(local_bash_path=local_bash_path, comp_env=comp_env, new_file=True,
                                                flags=['czvf'],
                                                include_list=include_list)

        if new_file:
            new_file_name = f'{structure.name}_nltha_case_{analysis_case}_' \
                            f'stampede_download_post_{Utility.get_curr_time_stamp()}.tar'
        else:
            new_file_name = f'{structure.name}_nltha_case_{analysis_case}_stampede_download_post.tar'

        os.replace(temp_file_name, new_file_name)
        return

    # ------------------------------------------------------------------------------------------------------------------

    def get_converged_analysis_count(self, analysis_case):
        conv_count, n_sim = self.recheck_status(analysis_case, print_status_file=False)
        print(f'{conv_count}/{n_sim} done!')
        return conv_count

    # ------------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
"""
Created on Sat Dec  7 18:32:34 2019

@author: Angshuman Deb
"""

import numpy as np
from scipy.stats import norm
from scipy.stats import multivariate_normal
from scipy.optimize import root_scalar
from pyDOE2 import lhs
import sys


class MultivariateNataf:

    def __init__(self):
        pass

    def __call__(self, prob_dist_list, corr_matrix, corr_matrix_z=None):
        return FrozenMultivariateNataf(self, prob_dist_list, corr_matrix, corr_matrix_z)

    def pdf(self, x, prob_dist_list, corr_matrix, corr_matrix_z=None):
        return FrozenMultivariateNataf(self, prob_dist_list, corr_matrix, corr_matrix_z).pdf(x)

    def cdf(self, x, prob_dist_list, corr_matrix, corr_matrix_z=None):
        return FrozenMultivariateNataf(self, prob_dist_list, corr_matrix, corr_matrix_z).cdf(x)

    def rvs(self, prob_dist_list, corr_matrix, corr_matrix_z=None, size=None, random_state=None, method='mcs'):
        return FrozenMultivariateNataf(self, prob_dist_list, corr_matrix, corr_matrix_z).rvs(size=size,
                                                                                             random_state=random_state,
                                                                                             method=method)


multivariate_nataf = MultivariateNataf()


class FrozenMultivariateNataf:

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, dist, prob_dist_list, corr_matrix, corr_matrix_z=None):
        self._dist = dist
        self.prob_dist_list = prob_dist_list
        self.ndim = len(self.prob_dist_list)
        self._k = 8
        self._mean = np.array([prob_dist.mean() for prob_dist in self.prob_dist_list])
        self._std = np.array([prob_dist.std() for prob_dist in self.prob_dist_list])

        if corr_matrix.shape[0] != corr_matrix.shape[1] or corr_matrix.shape[0] != self.ndim:
            raise ValueError("Invalid corr_matrix!")
        if len(corr_matrix.shape) > 2:
            # implies each face for components of mixture
            temp = [prob_dist.dist.name for prob_dist in self.prob_dist_list]
            if temp != ['mixture'] * self.ndim:
                raise ValueError("Expecting all prob dists for this type of corr_matrix input to be mixtures!")
            temp = [prob_dist.mixture_size for prob_dist in self.prob_dist_list]
            if temp[1:] != temp[:-1]:
                raise ValueError("Expecting all mixtures for this type of corr_matrix input to have same size!")
            if temp != [corr_matrix.shape[2]] * self.ndim:
                raise ValueError(f"Expecting all mixtures for this type of corr_matrix input to have size = "
                                 f"{corr_matrix.shape[2]}!")
            temp = np.array([prob_dist.weights for prob_dist in self.prob_dist_list])
            if (temp != temp[0]).any():
                raise ValueError("Expecting all mixtures for this type of corr_matrix input to have "
                                 "equal weights!")

            # corr_matrix : (n_rv, n_rv, n_s)
            # total_mean : (n_rv,)
            total_mean = np.array([prob_dist.mean() for prob_dist in self.prob_dist_list])
            # total_sig : (n_rv,)
            total_sig = np.array([prob_dist.std() for prob_dist in self.prob_dist_list])
            # mean_all : (n_s, n_rv)
            mean_all = np.array([prob_dist.mean_each for prob_dist in self.prob_dist_list]).T
            # sig_all : (n_s, n_rv)
            sig_all = np.array([prob_dist.std_each for prob_dist in self.prob_dist_list]).T
            # weights : (n_s,)
            weights = self.prob_dist_list[0].weights
            self.corr_matrix = FrozenMultivariateNataf.corr_clipped(
                FrozenMultivariateNataf.get_total_correlation(corr_matrix,
                                                              total_mean,
                                                              total_sig,
                                                              mean_all,
                                                              sig_all,
                                                              weights)
            )
        else:
            self.corr_matrix = FrozenMultivariateNataf.corr_clipped(corr_matrix)

        temp = np.diag(self._std)
        self._covariance = temp @ self.corr_matrix @ temp

        if corr_matrix_z is None:
            self.corr_matrix_z = FrozenMultivariateNataf.corr_clipped(
                FrozenMultivariateNataf.get_corr_matrix_z(self.prob_dist_list, self.corr_matrix, self._k))
        else:
            self.corr_matrix_z = FrozenMultivariateNataf.corr_clipped(corr_matrix_z)

        eig_val, eig_vec = np.linalg.eig(self.corr_matrix_z)
        self.eig_val = np.real(eig_val)
        self.eig_vec = np.real(eig_vec)

    ####################################################################################################################
    # Public functionalities
    ####################################################################################################################

    def mean(self):
        return self._mean

    # ------------------------------------------------------------------------------------------------------------------

    def std(self):
        return self._std

    # ------------------------------------------------------------------------------------------------------------------

    def covariance(self):
        return self._covariance

    # ------------------------------------------------------------------------------------------------------------------

    def pdf(self, x):
        return self._nataf_pdf(x)

    # ------------------------------------------------------------------------------------------------------------------

    def cdf(self, x):
        return self._nataf_cdf(x)

    # ------------------------------------------------------------------------------------------------------------------

    def rvs(self, size=None, random_state=None, method='mcs'):
        if size is None:
            size = 1
        try:
            size = np.array(size).item()
        except ValueError:
            pass
        rvs_temp = self.sampler(size, random_state, method)
        if size == 1 and self.ndim == 1:
            return rvs_temp.item()
        if size == 1 and self.ndim > 1:
            return rvs_temp.flatten()
        if np.isscalar(size) and self.ndim == 1:
            return rvs_temp.flatten()
        return rvs_temp.squeeze()

    # ------------------------------------------------------------------------------------------------------------------

    def _nataf_pdf(self, x):
        # x : (i, j, k, ..., ndim)
        z = np.ones(x.shape)
        term_1 = np.ones(x.shape[:-1])
        term_3 = np.ones(term_1.shape)
        for itr in range(self.ndim):
            term_1 *= self.prob_dist_list[itr].pdf(x[..., itr])
            temp = self.prob_dist_list[itr].cdf(x[..., itr])
            temp[temp > 1.0] = 1.0
            z_itr = norm.ppf(temp)
            z_itr[z_itr >= self._k] = self._k
            z_itr[z_itr <= -self._k] = -self._k
            term_3 *= norm.pdf(z_itr)
            z[..., itr] = z_itr
        mvn = multivariate_normal(mean=np.zeros(self.ndim),
                                  cov=self.corr_matrix_z)
        term_2 = mvn.pdf(z)
        return term_1 * term_2 / term_3

    # ------------------------------------------------------------------------------------------------------------------

    def _nataf_cdf(self, x):
        # x : (i, j, k, ..., ndim)
        z = np.ones(x.shape)
        for itr in range(self.ndim):
            temp = self.prob_dist_list[itr].cdf(x[..., itr])
            temp[temp > 1.0] = 1.0
            z_itr = norm.ppf(temp)
            z_itr[z_itr >= self._k] = self._k
            z_itr[z_itr <= -self._k] = -self._k
            z[..., itr] = z_itr
        mvn = multivariate_normal(mean=np.zeros(self.ndim),
                                  cov=self.corr_matrix_z)
        return mvn.cdf(z)

    # ------------------------------------------------------------------------------------------------------------------

    def sampler(self, size, random_state, method, return_all=False):
        if method == 'lhs':
            if not np.isscalar(size):
                n_samples = size[0]
                if len(size) == 2:
                    n_set = size[1]
                elif len(size) > 2:
                    raise ValueError("Non-scalar size for this type of sampling method "
                                     "should not be of length more than 2 and should represent "
                                     "(number of realizations of random vector, number of sets of realizations)!")
                else:
                    n_set = 1
            else:
                n_samples = size
                n_set = 1
        else:
            n_samples = None
            n_set = None

        if np.isscalar(size):
            x = np.zeros((size, self.ndim))
        else:
            x = np.zeros((*size, self.ndim))

        if method == 'lhs':
            for itr_set in range(n_set):
                if random_state is None:
                    rng_seed = None
                else:
                    if itr_set > 0:
                        rng_seed = int(''.join([str(i) for i in [random_state, itr_set]]))
                    else:
                        rng_seed = random_state
                criterion = "m"
                temp = lhs(self.ndim, n_samples, random_state=rng_seed, criterion=criterion)
                if len(x.shape) == 3:
                    x[:, itr_set, :] = temp
                else:
                    x = temp

        elif method == 'mcs':
            np.random.seed(random_state)
            x = np.random.uniform(0.0, 1.0, size=x.shape)
        else:
            return

        if return_all:
            return self.transform_samples(x)
        else:
            return self.transform_samples(x)[-1]

    def transform_samples(self, x):
        transformed_samples = np.zeros_like(x)
        independent_samples = norm.ppf(x)
        dependent_samples = independent_samples @ np.diag(np.sqrt(self.eig_val)) @ self.eig_vec.T
        uniform_dependent_samples = norm.cdf(dependent_samples)

        for itr_dim in range(self.ndim):
            transformed_samples[..., itr_dim] = self.prob_dist_list[itr_dim].ppf(
                uniform_dependent_samples[..., itr_dim])
        return x, independent_samples, dependent_samples, uniform_dependent_samples, transformed_samples

    def get_2d_lhs_grid(self, n, n_cont):
        eps = sys.float_info.epsilon
        x_vec = np.linspace(eps, 1 - eps, n + 1)
        x_vec_grid = np.linspace(eps, 1 - eps, n_cont)
        for itr in range(len(x_vec)):
            if not np.any(x_vec_grid == x_vec[itr]):
                x_vec_grid = np.hstack(
                    (x_vec_grid[x_vec_grid < x_vec[itr]],
                     x_vec[itr],
                     x_vec_grid[x_vec_grid > x_vec[itr]]))
        ind = [np.where(x_vec_grid == item)[0][0] for item in x_vec]
        xx, yy = np.meshgrid(x_vec_grid, x_vec_grid)
        x = np.zeros((*xx.shape, self.ndim))
        x[..., 0] = xx
        x[..., 1] = yy
        return self.transform_samples(x), ind

    ####################################################################################################################
    # Static methods
    ####################################################################################################################

    @staticmethod
    def _nataf_func_solve_1(x, *args):
        rho_ij_prime = x
        rho_ij, term_1, z, z_i, z_j = args
        n = len(z)
        mean = np.array([0, 0])
        cov = np.array([[1, rho_ij_prime], [rho_ij_prime, 1]])
        mvn = multivariate_normal(mean=mean, cov=cov)
        term_2 = mvn.pdf(np.column_stack([z_i.flatten(), z_j.flatten()])).reshape(n, n)
        integrand = term_1 * term_2
        integral = np.trapz(np.trapz(integrand, z, axis=1), z)
        return integral - rho_ij

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def _nataf_func_solve_2(x, *args):
        rho_ij_prime = x
        rho_ij, term_1, eta, xi = args
        mvn = multivariate_normal(mean=np.array([0, 0]), cov=np.array([[1, rho_ij_prime], [rho_ij_prime, 1]]))
        integral = np.sum(term_1 * mvn.pdf(np.column_stack([eta, xi])))
        return integral - rho_ij

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def corr_clipped(corr, threshold=1e-15):
        x_new, clipped = FrozenMultivariateNataf.clip_evals(corr, value=threshold)
        if not clipped:
            return corr

        # cov2corr
        x_std = np.sqrt(np.diag(x_new))
        x_new = x_new / np.outer(x_std, x_std)
        return x_new

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def clip_evals(x, value=0.0):
        evals, evecs = np.linalg.eigh(x)
        clipped = np.any(evals < value)
        x_new = np.dot(evecs * np.maximum(evals, value), evecs.T)
        return x_new, clipped

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_total_correlation(corr_matrix, total_mean, total_sig, mean_all, sig_all, weights):
        # corr_matrix : (n_rv, n_rv, n_s)
        # total_mean : (n_rv,)
        # total_sig : (n_rv,)
        # mean_all : (n_s, n_rv)
        # sig_all : (n_s, n_rv)
        # weights : (n_s,)

        weights = weights[np.newaxis, np.newaxis, :]
        term_1 = corr_matrix * np.einsum('ki,kj->ijk', sig_all, sig_all)
        temp = mean_all - total_mean[np.newaxis, :]
        term_2 = np.einsum('ki,kj->ijk', temp, temp)
        total_covariance = np.sum((term_1 + term_2) * weights, axis=2)
        total_corr = total_covariance / np.einsum('i,j->ij', total_sig, total_sig)
        # zero_corr_mask = corr_matrix[:, :, 0] == 0
        # total_corr[zero_corr_mask] = 0.0
        return total_corr

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_corr_matrix_z(prob_dist_list, corr_matrix, k):
        n_dim = len(prob_dist_list)
        corr_matrix_z = np.eye(n_dim)

        # method 1
        # z = np.linspace(-6.5, 6.5, 100)
        # # z = np.linspace(-8, 8, 1024)
        # z_i, z_j = np.meshgrid(z, z)
        #

        # method 2
        n = 100
        z_max = k
        z_min = -z_max
        points, weights = np.polynomial.legendre.leggauss(n)
        points = - (0.5 * (points + 1) * (z_max - z_min) + z_min)
        weights = weights * (0.5 * (z_max - z_min))
        xi = np.tile(points, [n, 1])
        xi = xi.flatten(order='F')
        eta = np.tile(points, n)
        first = np.tile(weights, n)
        first = np.reshape(first, [n, n])
        second = np.transpose(first)
        weights2d = first * second
        w2d = weights2d.flatten()
        #

        for i in range(n_dim):
            for j in range(n_dim):
                if i > j and corr_matrix[i, j] != 0.0:
                    mu_i = prob_dist_list[i].mean()
                    mu_j = prob_dist_list[j].mean()
                    sig_i = prob_dist_list[i].std()
                    sig_j = prob_dist_list[j].std()

                    # method 1
                    # x_i = prob_dist_list[i].ppf(norm.cdf(z_i))
                    # x_j = prob_dist_list[j].ppf(norm.cdf(z_j))
                    # term_1 = ((x_i - mu_i) / sig_i) * ((x_j - mu_j) / sig_j)
                    # args_1 = (corr_matrix[i, j], term_1, z, z_i, z_j)
                    #

                    # method 2
                    x_i = prob_dist_list[i].ppf(norm.cdf(points))  # x_i = ...(eta), do to x_i what you did to eta
                    x_j = prob_dist_list[j].ppf(norm.cdf(points))  # x_j = ...(xi), do to x_j what you did to xi
                    x_i = np.tile(x_i, n)
                    x_j = np.tile(x_j, [n, 1]).flatten(order='F')
                    term_1 = ((x_i - mu_i) / sig_i) * ((x_j - mu_j) / sig_j) * w2d
                    args_2 = (corr_matrix[i, j], term_1, eta, xi)
                    #

                    solve = True
                    if prob_dist_list[i].dist.name in ['norm', 'truncnorm']:
                        if prob_dist_list[j].dist.name in ['norm', 'truncnorm']:
                            corr_matrix_z[i, j] = corr_matrix[i, j]
                            solve = False
                        if prob_dist_list[j].dist.name == 'lognorm':
                            cov_j = sig_j / np.abs(mu_j)
                            corr_matrix_z[i, j] = cov_j / np.sqrt(np.log(1 + (cov_j ** 2))) * corr_matrix[i, j]
                            solve = False
                    if prob_dist_list[i].dist.name in ['lognorm']:
                        if prob_dist_list[j].dist.name in ['norm', 'truncnorm']:
                            cov_i = sig_i / np.abs(mu_i)
                            corr_matrix_z[i, j] = cov_i / np.sqrt(np.log(1 + (cov_i ** 2))) * corr_matrix[i, j]
                            solve = False
                        if prob_dist_list[j].dist.name == 'lognorm':
                            cov_i = sig_i / np.abs(mu_i)
                            cov_j = sig_j / np.abs(mu_j)
                            fac_num = np.log(1 + corr_matrix[i, j] * cov_i * cov_j)
                            fac_den = corr_matrix[i, j] * np.sqrt(np.log(1 + (cov_i ** 2)) * np.log(1 + (cov_j ** 2)))
                            corr_matrix_z[i, j] = (fac_num / fac_den) * corr_matrix[i, j]
                            solve = False
                    if solve:
                        # sol_1 = root_scalar(FrozenMultivariateNataf._nataf_func_solve_1, args=args_1,
                        #                     method='secant', x0=args_1[0], x1=args_1[0] * 0.9)
                        sol_2 = root_scalar(FrozenMultivariateNataf._nataf_func_solve_2, args=args_2,
                                            method='secant', x0=args_2[0], x1=args_2[0] * 0.9)
                        corr_matrix_z[i, j] = sol_2.root
                    corr_matrix_z[j, i] = corr_matrix_z[i, j]
        return corr_matrix_z

    # ------------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
"""
Abstract class : StructuralAnalysisPlatform
Derived class : OpenSeesTcl
Derived class : OpenSeesPy

@author: Angshuman Deb
"""

import json
import os.path
import platform
import subprocess
import numpy as np
from .utility import Utility
from abc import ABC, abstractmethod


class StructuralAnalysisPlatform(ABC):
    def __init__(self, model_files_path):
        self.model_files_path = Utility.get_path(model_files_path)
        self.file_ext = None
        self.nltha_rec_dir_name = 'NLTHA_Recorder_Files'
        self.model_info_dir_name = 'Model_Info_Files'

    @abstractmethod
    def bat_file(self, dest_path):
        pass

    @abstractmethod
    def write_additional_recorders(self, recorder_info, dest_path, file_open_mode):
        pass

    @abstractmethod
    def open_edp_recorder_file(self, work_dir_path, rec_gen_file_open_mode, rec_save_dir_path):
        pass

    @abstractmethod
    def write_evaluate_edp_and_helpers(self, edp_tag, max_what, resp, resp_num):
        pass

    @abstractmethod
    def write_evaluate_edps(self, work_dir_path, edp, for_which, gen_rec, mode):
        pass

    @abstractmethod
    def write_run_prelim_analysis(self, work_dir_path, num_modes):
        pass

    @abstractmethod
    def write_run_nltha(self, target_work_dir_path, work_dir_path_prelim_analysis, **kwargs):
        pass

    @abstractmethod
    def get_prelim_analysis_files(self):
        pass

    @abstractmethod
    def get_nltha_files(self):
        pass

    @abstractmethod
    def run_prelim_analysis(self, target_work_dir_path, comp_env):
        pass

    @abstractmethod
    def run_nltha(self, target_work_dir_path, comp_env):
        pass

    @abstractmethod
    def run_file(self, target_work_dir_path, run_file_name, comp_env):
        pass

    @abstractmethod
    def perform_mat_test(self, mat_data_dir_path, mat_tag, input_data, num_incr):
        pass

    @staticmethod
    @abstractmethod
    def write_model_param_vals(work_dir_path, name_value_pairs):
        pass

    @staticmethod
    @abstractmethod
    def write_model_attributes(work_dir_path, name_value_pairs):
        pass

    @staticmethod
    @abstractmethod
    def write_damping_file(damping_model, work_dir_path, **kwargs):
        pass

    @staticmethod
    @abstractmethod
    def get_stress_strain_recorder_command(recorder_file_path, ele_tag, sec_num, y, z, mat_tag):
        pass

    @staticmethod
    @abstractmethod
    def get_elem_deformation_recorder_command(recorder_file_path, ele_tag):
        pass

    @staticmethod
    @abstractmethod
    def get_recorder_files_to_read(rec_gen_file_path):
        pass


class OpenSeesTcl(StructuralAnalysisPlatform):
    def __init__(self, model_files_path, local_opensees_path):
        super().__init__(model_files_path)
        self.local_opensees_path = Utility.get_path(local_opensees_path)
        self.file_ext = '.tcl'

    def bat_file(self, dest_path):
        # OpenSees.bat
        if platform.system() == 'Windows':
            with open(Utility.get_path(dest_path, 'OpenSees.bat'), 'w') as fid:
                fid.write(f'"{self.local_opensees_path}"')

    def write_additional_recorders(self, recorder_info, dest_path, file_open_mode):
        element_response_list = recorder_info.get('element_response_list', [])
        element_section_response_list = recorder_info.get('element_section_response_list', [])
        element_section_fiber_response_list = recorder_info.get('element_section_fiber_response_list', [])
        node_response_list = recorder_info.get('node_response_list', [])
        addnl_file_list = recorder_info.get('addnl_file_list', [])

        if 'a' in file_open_mode:
            with open(Utility.get_path(dest_path, 'generate_additional_recorders.tcl'), 'r') as fid:
                lines = fid.readlines()
                if not lines[-1].endswith('\n'):
                    lines[-1] = lines[-1] + '\n'
        else:
            lines = []

        nltha_dir = self.nltha_rec_dir_name
        for item in element_response_list:
            ele_tag, what = item
            file_name = Utility.get_path(nltha_dir, f'element_response_elem_{ele_tag}_resp_{what}.txt')
            recorder_line = f'recorder Element -file "{file_name}" -time -ele {ele_tag} {what}\n'
            lines.append(recorder_line)

        for item in element_section_response_list:
            ele_tag, sec_num, what = item
            file_name = Utility.get_path(nltha_dir, f'element_section_response_elem_{ele_tag}_secnum_{sec_num}_resp_{what}.txt')
            recorder_line = f'recorder Element -file "{file_name}" -time -ele {ele_tag} section {sec_num} {what}\n'
            lines.append(recorder_line)

        for item in element_section_fiber_response_list:
            ele_tag, sec_num, fib_num, y, z, mat_tag = item
            file_name = Utility.get_path(nltha_dir, f'element_section_fiber_response_elem_{ele_tag}_secnum_{sec_num}_fib_{fib_num}_mat_{mat_tag}_resp_stressStrain.txt')
            recorder_line = f'recorder Element -file "{file_name}" -time -ele {ele_tag} section {sec_num} fiber {y} {z} {mat_tag} stressStrain\n'
            lines.append(recorder_line)

        for item in node_response_list:
            node_tag, dof, what = item
            file_name = Utility.get_path(nltha_dir, f'node_response_node_{node_tag}_dof_{dof}_resp_{what}.txt')
            recorder_line = f'recorder Node -file "{file_name}" -time -node {node_tag} -dof {dof} {what}\n'
            lines.append(recorder_line)

        for item in addnl_file_list:
            item = item.split('.')[0] + self.file_ext
            lines.append(f'source "{item}"\n')

        with open(Utility.get_path(dest_path, 'generate_additional_recorders.tcl'), 'w') as fid:
            fid.writelines(lines)

    def open_edp_recorder_file(self, work_dir_path, rec_gen_file_open_mode, rec_save_dir_path):
        rec_gen_file_fid = open(Utility.get_path(work_dir_path, 'generate_edp_recorders' + self.file_ext), rec_gen_file_open_mode)
        new_recorder_line = f'file mkdir "{rec_save_dir_path}";\n'
        rec_gen_file_fid.write(new_recorder_line)
        return rec_gen_file_fid

    def write_evaluate_edp_and_helpers(self, edp_tag, max_what, resp, resp_num):
        model_files_path = self.model_files_path
        lines = list()
        lines.append('proc max_list list {\n')
        lines.append('    set index 0\n')
        lines.append('    set max_index $index\n')
        lines.append('    set max_val [lindex $list 0]\n')
        lines.append('    foreach val $list {\n')
        lines.append('        if {$val > $max_val} {\n')
        lines.append('            set max_index $index\n')
        lines.append('            set max_val $val\n')
        lines.append('        }\n')
        lines.append('        incr index\n')
        lines.append('    }\n')
        lines.append('    return [list $max_val $max_index]\n')
        lines.append('}\n')
        max_list_file_path = Utility.get_path(model_files_path, 'max_list.tcl')
        with open(max_list_file_path, 'w') as fid:
            fid.writelines(lines)

        if max_what in ['compression', 'range']:
            lines = list()
            lines.append('proc min_list list {\n')
            lines.append('    set index 0\n')
            lines.append('    set min_index $index\n')
            lines.append('    set min_val [lindex $list 0]\n')
            lines.append('    foreach val $list {\n')
            lines.append('        if {$val < $min_val} {\n')
            lines.append('            set min_index $index\n')
            lines.append('            set min_val $val\n')
            lines.append('        }\n')
            lines.append('        incr index\n')
            lines.append('    }\n')
            lines.append('    return [list $min_val $min_index]\n')
            lines.append('}\n')
            min_list_file_path = Utility.get_path(model_files_path, 'min_list.tcl')
            with open(min_list_file_path, 'w') as fid:
                fid.writelines(lines)

        lines = list()
        lines.append(f'proc get_{resp}_vals file_path {{\n')
        lines.append(f'    set data_fid [open $file_path "r"]\n')
        lines.append(f'    set data [read $data_fid]\n')
        lines.append(f'    close $data_fid\n')
        lines.append(f'    set data_new [split $data "\\n"]\n')
        lines.append(f'    set {resp}_vals {{}}\n')
        lines.append(f'    for {{set k 0}} {{$k <= [expr [llength $data_new] - 2]}} {{incr k 1}} {{\n')
        lines.append(f'        set data_t [lindex $data_new $k]\n')
        lines.append(f'        lappend {resp}_vals [lindex $data_t [expr {resp_num}]]\n')
        lines.append(f'    }}\n')
        lines.append(f'    return ${resp}_vals\n')
        lines.append(f'}}\n')
        get_resp_vals_file_path = Utility.get_path(model_files_path, f'get_{resp}_vals.tcl')
        with open(get_resp_vals_file_path, 'w') as fid:
            fid.writelines(lines)

        lines = list()
        if max_what == 'tension':
            lines.append(f'proc evaluate_edp_{edp_tag} {resp}_vals {{\n')
            lines.append(f'    set temp [max_list ${resp}_vals]\n')
            lines.append(f'    return [lindex $temp 0]\n')
            lines.append(f'}}\n')
        if max_what == 'compression':
            lines.append(f'proc evaluate_edp_{edp_tag} {resp}_vals {{\n')
            lines.append(f'    set temp [min_list ${resp}_vals]\n')
            lines.append(f'    return [expr -([lindex $temp 0])]\n')
            lines.append(f'}}\n')
        if max_what == 'range':
            lines.append(f'proc evaluate_edp_{edp_tag} {resp}_vals {{\n')
            lines.append(f'    set temp [max_list ${resp}_vals]\n')
            lines.append(f'    set max_{resp} [lindex $temp 0]\n')
            lines.append(f'    set {resp}_vals [lrange ${resp}_vals [lindex $temp 1] end]\n')
            lines.append(f'    set temp [min_list ${resp}_vals]\n')
            lines.append(f'    set min_{resp} [lindex $temp 0]\n')
            lines.append(f'    return [expr $max_{resp} - $min_{resp}]\n')
            lines.append(f'}}\n')

        eval_file_path = Utility.get_path(model_files_path, f'evaluate_edp_{edp_tag}.tcl')
        with open(eval_file_path, 'w') as fid:
            fid.writelines(lines)

    def write_evaluate_edps(self, work_dir_path, edp, for_which, gen_rec, mode):
        rec_gen_file_path = Utility.get_path(work_dir_path, 'generate_edp_recorders' + self.file_ext)
        eval_edp_file_path = Utility.get_path(work_dir_path, 'evaluate_edps' + self.file_ext)

        recorder_files_to_read = OpenSeesTcl.get_recorder_files_to_read(rec_gen_file_path)
        if edp.recorder_file_storage == 'shared':
            recorder_files_to_read = [file for file in recorder_files_to_read
                                      if f'/EDP/' in file]
        else:
            recorder_files_to_read = [file for file in recorder_files_to_read
                                      if f'/EDP_{edp.tag}/' in file]

        lines = list()
        lines.append(f'source "{self.model_files_path}/evaluate_edp_{edp.tag}.tcl"\n')
        lines.append(f'file mkdir "EDP_{edp.tag}_Results"\n')
        if gen_rec:
            lines.append(f'source "{self.model_files_path}/get_{edp.resp}_vals.tcl"\n')
            lines.append(f'source "{self.model_files_path}/max_list.tcl"\n')
        if edp.max_what in ['compression', 'range']:
            lines.append(f'source "{self.model_files_path}/min_list.tcl"\n')

        edp_strings = edp.get_edp_strings(for_which)
        for edp_string in edp_strings:
            edp_var_name = f'edp_{edp.tag}_' + edp_string
            files_to_eval = [file for file in recorder_files_to_read if
                             edp_string in file]

            var_names_to_eval = [os.path.basename(file).split('.')[0] for file in files_to_eval]
            if gen_rec:
                for (var_name, file) in zip(var_names_to_eval, files_to_eval):
                    lines.append(f'set {var_name} [get_{edp.resp}_vals "{file}"]\n')

            lines.append(f'set {edp_var_name} {{0}}\n')
            for var_name in var_names_to_eval:
                lines.append(f'lappend {edp_var_name} [evaluate_edp_{edp.tag} ${var_name}]\n')

            lines.append(f'set temp [max_list ${edp_var_name}]\n')
            lines.append(f'set {edp_var_name} [lindex $temp 0]\n')
            lines.append(f'set {edp_var_name} [expr ${edp_var_name} / {edp.normalize_with}]\n')
            lines.append(f'set edp_file [open "EDP_{edp.tag}_Results/{edp_var_name}.txt" "w"]\n')
            lines.append(f'puts $edp_file ${edp_var_name}\n')
            lines.append(f'close $edp_file\n')

        with open(eval_edp_file_path, mode) as eval_file_fid:
            eval_file_fid.writelines(lines)
        return

    def write_run_prelim_analysis(self, work_dir_path, num_modes):
        model_files_path = self.model_files_path
        lines = ['wipe;\n',
                 f'set model_files_path "{model_files_path}";\n',
                 f'set model_info_directory "{self.model_info_dir_name}";\n',
                 'file mkdir $model_info_directory;\n',
                 'set write_model_files 1;\n',
                 f'set modes [expr int({num_modes})];\n',
                 'source "model_param_vals.tcl";\n',
                 'source "model_attributes.tcl";\n',
                 'source "$model_files_path/main.tcl";\n',
                 'source "$model_files_path/analysis_gravity.tcl";\n',
                 'source "$model_files_path/analysis_prelim.tcl";\n',
                 'wipe;\n']
        with open(Utility.get_path(work_dir_path, 'run_prelim_analysis.tcl'), 'w') as fid:
            fid.writelines(lines)

    def write_run_nltha(self, target_work_dir_path, work_dir_path_prelim_analysis, **kwargs):
        write_model_files = kwargs.get('write_model_files', 0)
        generate_additional_recorders = kwargs.get('generate_additional_recorders', False)
        delete_rec = kwargs.get('delete_rec', True)
        lines = [
            'wipe;\n',
            'set analysis_status_file [open "NLTHA_STATUS.txt" "w"];\n',
            'puts $analysis_status_file "FAIL";\n',
            'close $analysis_status_file;\n',
            f'set model_files_path "{self.model_files_path}";\n',
            f'set work_dir_path_prelim_analysis "{work_dir_path_prelim_analysis}";\n'
        ]
        if write_model_files == 0:
            lines += [
                'set write_model_files 0;\n',
            ]
        else:
            lines += [
                'set write_model_files 1;\n',
                f'set model_info_directory "{self.model_info_dir_name}";\n',
                'file mkdir $model_info_directory;\n',
            ]
        lines += [
            f'source "$work_dir_path_prelim_analysis/model_param_vals.tcl";\n',
            f'source "$work_dir_path_prelim_analysis/model_attributes.tcl";\n',
            'source "$model_files_path/main.tcl";\n',
            'source "$model_files_path/analysis_gravity.tcl";\n',
            'source "$work_dir_path_prelim_analysis/generate_damping.tcl";\n',
            'source "generate_edp_recorders.tcl";\n',
        ]
        if generate_additional_recorders:
            lines += [
                'source "generate_additional_recorders.tcl";\n'
            ]
        lines += [
            f'set gm_file_loc ".";\n',
            'source "$model_files_path/analysis_time_hist.tcl";\n',
            'remove recorders;\n',
            'if {$NLTHA_OK == 1} {\n',
            '    source "evaluate_edps.tcl";\n',
            '    set analysis_status_file [open "NLTHA_STATUS.txt" "w"];\n',
            '    puts $analysis_status_file "SUCCESS"\n',
            '    close $analysis_status_file;\n'
            '}\n',
        ]
        if delete_rec:
            lines += [
                f'file delete -force -- "{self.nltha_rec_dir_name}";\n',
            ]
        lines += [
            'wipe;\n'
        ]
        with open(Utility.get_path(target_work_dir_path, 'run_nltha.tcl'), 'w') as fid:
            fid.writelines(lines)

    def get_prelim_analysis_files(self):
        return [item + self.file_ext for item in ['generate_damping', 'model_param_vals', 'model_attributes']]

    def get_nltha_files(self):
        return ['evaluate_edps.tcl', 'generate_edp_recorders.tcl']

    def run_prelim_analysis(self, target_work_dir_path, comp_env):
        self.run_file(target_work_dir_path, 'run_prelim_analysis.tcl', comp_env)

    def run_nltha(self, target_work_dir_path, comp_env):
        self.run_file(target_work_dir_path, 'run_nltha.tcl', comp_env)

    def run_file(self, target_work_dir_path, run_file_name, comp_env):
        curr_dir = Utility.get_path(os.getcwd())
        os.chdir(target_work_dir_path)
        if comp_env == 'local' and platform.system() == 'Windows':
            subprocess.call(
                # f"\"{self.local_opensees_path}\" \"{run_file_name}\"", shell=True,
                [self.local_opensees_path, run_file_name], shell=True,
                # stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            )
        else:
            os.system(f"\"{self.local_opensees_path}\" \"{run_file_name}\"")
        if os.path.isfile('gmon.out'):
            os.remove('gmon.out')
        os.chdir(curr_dir)

    def perform_mat_test(self, mat_data_dir_path, mat_tag, input_data, num_incr):
        material_data_file_path = Utility.get_path(mat_data_dir_path, 'material_data.txt')
        with open(material_data_file_path, 'r') as fid:
            lines = fid.readlines()
            lines = [line for line in lines if not line.strip().isspace() and line.strip()]
            material_data = ' '.join([line for line in lines if line.split()[2] == str(mat_tag)][0].split()) + '\n'

        lines = list()
        lines.append(f'wipe;\n')
        lines.append(f'model testUniaxial;\n')
        lines.append(f'set mat_tag {mat_tag};\n')
        lines.append(f'set strain_history {{{input_data}}};\n')
        lines.append(f'set file_out "hysteresis_mat_$mat_tag.txt";\n')
        lines.append(f'set out [open $file_out w];\n')
        lines.append(material_data)

        lines += ['uniaxialTest $mat_tag;\n',
                  'set strain 0.0;\n',
                  'set i_time 0;\n',
                  'set strain [expr $strain];\n',
                  'strainUniaxialTest $strain;\n',
                  'set stress [stressUniaxialTest];\n',
                  'set tangent [tangUniaxialTest];\n',
                  'set i_time [expr $i_time+1];\n',
                  'puts $out "$strain $stress";\n',
                  f'set num_incr {int(num_incr)};\n',
                  'foreach {strain_extreme_val} $strain_history {\n',
                  '    set strain_incr [expr ($strain_extreme_val - $strain)/$num_incr];\n',
                  '    for {set i 0} {$i < $num_incr} {incr i 1} {\n',
                  '	       set strain [expr $strain+$strain_incr];\n',
                  '        strainUniaxialTest $strain;\n',
                  '        set stress [stressUniaxialTest];\n',
                  '        set tangent [tangUniaxialTest];\n',
                  '        set i_time [expr $i_time+1];\n',
                  '        puts $out "$strain $stress";\n',
                  '    }\n',
                  '}\n',
                  'close $out;\n',
                  'puts "MATERIAL TESTER RAN SUCCESSFULLY!";\n',
                  'wipe;\n'
                  ]
        with open(Utility.get_path(mat_data_dir_path, 'mat_test.tcl'), 'w+') as fid:
            fid.writelines(lines)
        self.run_file(mat_data_dir_path, 'mat_test.tcl', 'local')
        out = Utility.read_numpy_array_from_txt_file(
            Utility.get_path(mat_data_dir_path, f'hysteresis_mat_{mat_tag}.txt'))
        os.remove(Utility.get_path(mat_data_dir_path, f'hysteresis_mat_{mat_tag}.txt'))
        os.remove(Utility.get_path(mat_data_dir_path, 'mat_test.tcl'))
        return out

    @staticmethod
    def write_model_param_vals(work_dir_path, name_value_pairs):
        OpenSeesTcl.write_tcl_input_params_file(work_dir_path, name_value_pairs, "model_param_vals.tcl")

    @staticmethod
    def write_model_attributes(work_dir_path, name_value_pairs):
        OpenSeesTcl.write_tcl_input_params_file(work_dir_path, name_value_pairs, "model_attributes.tcl")

    @staticmethod
    def write_damping_file(damping_model, work_dir_path, **kwargs):
        lines = []
        if damping_model == 'rayleigh_damping':
            alpha = kwargs['alpha']
            beta = kwargs['beta']
            lines.extend([f'rayleigh {alpha} 0. 0. {beta}\n'])
        if damping_model == 'modal_damping':
            n_modes = kwargs['n_modes']
            damp_ratios = kwargs['damp_ratios']
            lines.extend(
                [
                    f'eigen {int(n_modes)}\n',
                    f"modalDamping {' '.join(np.array(damp_ratios).astype(str))}\n"
                    f"wipeAnalysis\n"
                ]
            )
        with open(Utility.get_path(work_dir_path, 'generate_damping.tcl'), 'w') as fid:
            fid.writelines(lines)
        return

    @staticmethod
    def get_stress_strain_recorder_command(recorder_file_path, ele_tag, sec_num, y, z, mat_tag):
        new_recorder_line = f'recorder Element -file "{recorder_file_path}" -time -ele {ele_tag} ' \
                            f'section {sec_num} fiber {y} {z} {mat_tag} stressStrain\n'
        return new_recorder_line

    @staticmethod
    def get_elem_deformation_recorder_command(recorder_file_path, ele_tag):
        new_recorder_line = f'recorder Element -file "{recorder_file_path}" -time -ele {ele_tag} deformation;\n'
        return new_recorder_line

    @staticmethod
    def write_tcl_input_params_file(work_dir_path, name_value_pairs, filename):
        with open(Utility.get_path(work_dir_path, filename), 'w') as fid:
            for name, value in name_value_pairs:
                Utility.write_tcl_input_file(
                    fid, value, name
                )

    @staticmethod
    def get_recorder_files_to_read(rec_gen_file_path):
        with open(rec_gen_file_path, 'r') as fid:
            lines = fid.readlines()
        lines = [line.split() for line in lines if 'recorder' in line and '-file' in line]
        return [line[line.index('-file') + 1] for line in lines]


class OpenSeesPy(StructuralAnalysisPlatform):
    def __init__(self, model_files_path, local_python_path):
        super().__init__(model_files_path)
        self.local_python_path = Utility.get_path(local_python_path)
        self.file_ext = '.py'

    def bat_file(self, dest_path):
        pass

    def write_additional_recorders(self, recorder_info, dest_path, file_open_mode):
        element_response_list = recorder_info.get('element_response_list', [])
        element_section_response_list = recorder_info.get('element_section_response_list', [])
        element_section_fiber_response_list = recorder_info.get('element_section_fiber_response_list', [])
        node_response_list = recorder_info.get('node_response_list', [])
        addnl_file_list = recorder_info.get('addnl_file_list', [])

        if 'a' in file_open_mode:
            with open(Utility.get_path(dest_path, 'generate_additional_recorders.py'), 'r') as fid:
                lines = fid.readlines()
                if not lines[-1].endswith('\n'):
                    lines[-1] = lines[-1] + '\n'
        else:
            lines = [
                'import openseespy.opensees as ops\n',
                '\n',
                '\n',
                'def generate_additional_recorders():\n',
            ]

        importlib_line = 'import importlib.util\n'
        if len(addnl_file_list) > 0 and importlib_line not in lines:
            lines = [importlib_line, *lines]

        nltha_dir = self.nltha_rec_dir_name
        for item in element_response_list:
            ele_tag, what = item
            file_name = Utility.get_path(nltha_dir, f'element_response_elem_{ele_tag}_resp_{what}.txt')
            recorder_line = f'    ops.recorder(\'Element\', \'-file\', \'{file_name}\', \'-time\', \'-ele\', {ele_tag}, \'{what}\')\n'
            lines.append(recorder_line)

        for item in element_section_response_list:
            ele_tag, sec_num, what = item
            file_name = Utility.get_path(nltha_dir, f'element_section_response_elem_{ele_tag}_secnum_{sec_num}_resp_{what}.txt')
            recorder_line = f'    ops.recorder(\'Element\', \'-file\', \'{file_name}\', \'-time\', \'-ele\', {ele_tag}, \'section\', {sec_num}, \'{what}\')\n'
            lines.append(recorder_line)

        for item in element_section_fiber_response_list:
            ele_tag, sec_num, fib_num, y, z, mat_tag = item
            file_name = Utility.get_path(nltha_dir, f'element_section_fiber_response_elem_{ele_tag}_secnum_{sec_num}_fib_{fib_num}_mat_{mat_tag}_resp_stressStrain.txt')
            recorder_line = f'    ops.recorder(\'Element\', \'-file\', \'{file_name}\', \'-time\', \'-ele\', {ele_tag}, \'section\', {sec_num}, \'fiber\', {y}, {z}, {mat_tag}, \'stressStrain\')\n'
            lines.append(recorder_line)

        for item in node_response_list:
            node_tag, dof, what = item
            file_name = Utility.get_path(nltha_dir, f'node_response_node_{node_tag}_dof_{dof}_resp_{what}.txt')
            recorder_line = f'    ops.recorder(\'Node\', \'-file\', \'{file_name}\', \'-time\', \'-node\', {node_tag}, \'-dof\', {dof}, \'{what}\')\n'
            lines.append(recorder_line)

        for item in addnl_file_list:
            module_name = item.split('.')[0]
            lines.extend(OpenSeesPy.import_module(dest_path, module_name, module_name, tab='    '))
            lines.append(f'    {module_name}.{module_name}()\n')

        with open(Utility.get_path(dest_path, 'generate_additional_recorders.py'), 'w') as fid:
            fid.writelines(lines)

    def open_edp_recorder_file(self, work_dir_path, rec_gen_file_open_mode, rec_save_dir_path):
        rec_gen_file_fid = open(Utility.get_path(work_dir_path, 'generate_edp_recorders' + self.file_ext), rec_gen_file_open_mode)
        if 'w' in rec_gen_file_open_mode:
            lines = [
                f'import os\n',
                f'import openseespy.opensees as ops\n',
                f'\n\n',
                f'def generate_edp_recorders():\n',
                f'    os.makedirs(\'{rec_save_dir_path}\', exist_ok=True)\n',
            ]
        else:
            lines = [
                f'    os.makedirs(\'{rec_save_dir_path}\', exist_ok=True)\n',
            ]
        rec_gen_file_fid.writelines(lines)
        return rec_gen_file_fid

    def write_evaluate_edp_and_helpers(self, edp_tag, max_what, resp, resp_num):
        get_resp_vals = [
            f"def get_{resp}_vals(file_path):\n",
            f"    data = np.loadtxt(file_path, delimiter=' ')\n",
            f"    return data[:, {resp_num}]\n",
        ]

        edp_eval_func_name = f'evaluate_edp_{edp_tag}'

        max_list = [
            f"def {edp_eval_func_name}(x):\n",
            f"    return np.max(x)\n",
        ]

        abs_min_list = [
            f"def {edp_eval_func_name}(x):\n",
            f"    return np.abs(np.min(x))\n",
        ]

        max_range = [
            f"def {edp_eval_func_name}(x):\n",
            f"    max_x = np.max(x)\n",
            f"    max_x_i = np.argmax(x)\n",
            f"    x = x[max_x_i:]\n",
            f"    min_x_following = np.min(x)\n",
            f"    return max_x - min_x_following\n",
        ]

        model_files_path = self.model_files_path

        lines = ['import numpy as np\n\n\n'] + get_resp_vals
        get_resp_vals_file_path = Utility.get_path(model_files_path, f'get_{resp}_vals.py')
        with open(get_resp_vals_file_path, 'w') as fid:
            fid.writelines(lines)

        lines = ['import numpy as np\n\n\n']
        if max_what == 'tension':
            lines += max_list
        if max_what == 'compression':
            lines += abs_min_list
        if max_what == 'range':
            lines += max_range
        eval_file_path = Utility.get_path(model_files_path, f'{edp_eval_func_name}.py')
        with open(eval_file_path, 'w') as fid:
            fid.writelines(lines)

    def write_evaluate_edps(self, work_dir_path, edp, for_which, gen_rec, mode):
        rec_gen_file_path = Utility.get_path(work_dir_path, 'generate_edp_recorders' + self.file_ext)
        eval_edp_file_path = Utility.get_path(work_dir_path, 'evaluate_edps' + self.file_ext)

        recorder_files_to_read = OpenSeesPy.get_recorder_files_to_read(rec_gen_file_path)
        if edp.recorder_file_storage == 'shared':
            recorder_files_to_read = [file for file in recorder_files_to_read
                                      if f'/EDP/' in file]
        else:
            recorder_files_to_read = [file for file in recorder_files_to_read
                                      if f'/EDP_{edp.tag}/' in file]
        lines = list()
        if 'w' in mode:
            lines.extend(['import importlib.util\n', 'import os\n\n\n', 'def evaluate_edps():\n'])
            lines.extend(OpenSeesPy.import_module(self.model_files_path, 'ops_model', 'model', tab='    '))
        lines.extend(OpenSeesPy.import_module(self.model_files_path, f'evaluate_edp_{edp.tag}', f'edp_{edp.tag}', tab='    '))
        if gen_rec:
            lines.extend(OpenSeesPy.import_module(self.model_files_path, f'get_{edp.resp}_vals', f'resp_{edp.resp}', tab='    '))
        lines.append(f'    os.makedirs(\'EDP_{edp.tag}_Results\', exist_ok=True)\n')
        edp_strings = edp.get_edp_strings(for_which)
        if isinstance(edp.normalize_with, str):
            normalize_with = 'model.' + edp.normalize_with.strip('$')
        else:
            normalize_with = f'{edp.normalize_with}'
        for i, edp_string in enumerate(edp_strings):
            edp_var_name = f'edp_{edp.tag}_' + edp_string
            files_to_eval = [file for file in recorder_files_to_read if
                             edp_string in file]
            var_names_to_eval = [os.path.basename(file).split('.')[0] for file in files_to_eval]
            if gen_rec:
                for (var_name, file) in zip(var_names_to_eval, files_to_eval):
                    lines.append(f'    {var_name} = resp_{edp.resp}.get_{edp.resp}_vals("{file}")\n')
            lines.append(f'    {edp_var_name} = []\n')
            for var_name in var_names_to_eval:
                lines.append(f'    {edp_var_name}.append(edp_{edp.tag}.evaluate_edp_{edp.tag}({var_name}))\n')
            lines.append(f'    {edp_var_name} = max({edp_var_name}) / {normalize_with}\n')
            lines.append(f'    with open(\'EDP_{edp.tag}_Results/{edp_var_name}.txt\', \'w\') as edp_file:\n')
            lines.append(f'        edp_file.write(str({edp_var_name}) + \'\\n\')\n')

        with open(eval_edp_file_path, mode) as eval_file_fid:
            eval_file_fid.writelines(lines)
        return

    def write_run_prelim_analysis(self, work_dir_path, num_modes):
        model_files_path = self.model_files_path
        model_info_directory = Utility.get_path(work_dir_path, self.model_info_dir_name)
        module_name = 'ops_model'
        lines = [
            f'import os\n',
            f'import json\n',
            f'import importlib.util\n\n',
        ]
        lines.extend(OpenSeesPy.import_module(self.model_files_path, module_name, 'model'))
        lines.extend(
            [
                f'model_files_path = \'{model_files_path}\'\n',
                f'model_info_directory = \'{model_info_directory}\'\n',
                'os.makedirs(model_info_directory, exist_ok=True)\n',
            ]
        )
        lines.extend(OpenSeesPy.read_json('model_param_vals.json', 'model_param_vals'))
        lines.extend(OpenSeesPy.read_json('model_attributes.json', 'model_attributes'))
        names = [
            'model_files_path',
            'model_info_directory',
            'write_model_files',
            'modes',
            'model_param_vals',
            'model_attributes',
        ]
        values = [
            'model_files_path',
            'model_info_directory',
            1,
            num_modes,
            'model_param_vals',
            'model_attributes',
        ]
        lines.extend(OpenSeesPy.create_dict(names, values, 'model_input_dict'))
        lines.extend(
            [
                "model.main(model_input_dict)\n",
                'model.analysis_gravity(model_input_dict)\n',
                'model.analysis_prelim(model_input_dict)\n',
            ]
        )
        with open(Utility.get_path(work_dir_path, 'run_prelim_analysis.py'), 'w') as fid:
            fid.writelines(lines)

    def write_run_nltha(self, target_work_dir_path, work_dir_path_prelim_analysis, **kwargs):
        write_model_files = kwargs.get('write_model_files', 0)
        generate_additional_recorders = kwargs.get('generate_additional_recorders', False)
        delete_rec = kwargs.get('delete_rec', True)

        lines = [
            'import json\n',
            'import importlib.util\n',
            'import shutil\n',
            'import openseespy.opensees as ops\n\n',
            'with open("NLTHA_STATUS.txt", "w") as fid:\n',
            '    fid.write("FAIL\\n")\n',
        ]
        lines.extend(OpenSeesPy.import_module(self.model_files_path, 'ops_model', 'model'))
        lines.extend(OpenSeesPy.import_module(work_dir_path_prelim_analysis, 'generate_damping', 'damp'))
        lines.extend(OpenSeesPy.import_module('', 'generate_edp_recorders', 'edp_rec'))
        lines.extend(OpenSeesPy.import_module('', 'evaluate_edps', 'edp_eval'))
        lines.extend(OpenSeesPy.read_json(Utility.get_path(work_dir_path_prelim_analysis, 'model_param_vals.json'), 'model_param_vals'))
        lines.extend(OpenSeesPy.read_json(Utility.get_path(work_dir_path_prelim_analysis, 'model_attributes.json'), 'model_attributes'))
        names = [
            'write_model_files',
            'model_param_vals',
            'model_attributes',
            'gm_file_loc',
        ]
        values = [
            write_model_files,
            'model_param_vals',
            'model_attributes',
            '\'.\'',
        ]
        if write_model_files == 1:
            model_info_directory = Utility.get_path(target_work_dir_path, self.model_info_dir_name)
            lines = ['import os\n', *lines]
            lines += [
                f'model_info_directory = \'{model_info_directory}\'\n',
                'os.makedirs(model_info_directory, exist_ok=True)\n',
            ]
            names += ['model_info_directory']
            values += ['model_info_directory']
        lines.extend(OpenSeesPy.create_dict(names, values, 'model_input_dict'))
        lines.extend(
            [
                "model.main(model_input_dict)\n",
                'model.analysis_gravity(model_input_dict)\n',
                'damp.generate_damping()\n',
                'edp_rec.generate_edp_recorders()\n',
            ]
        )
        if generate_additional_recorders:
            lines.extend(OpenSeesPy.import_module(self.model_files_path, 'generate_additional_recorders', 'addnl_rec'))
            lines += [
                'addnl_rec.generate_additional_recorders()\n',
            ]

        lines += [
            'nltha_ok = model.analysis_time_hist(model_input_dict)\n',
            'ops.remove(\'recorders\')\n',
            'if nltha_ok:\n',
            '    edp_eval.evaluate_edps()\n',
            '    with open("NLTHA_STATUS.txt", "w") as fid:\n',
            '        fid.write("SUCCESS\\n")\n',
        ]
        if delete_rec:
            lines += [
                f'shutil.rmtree("{self.nltha_rec_dir_name}")\n',
            ]
        with open(Utility.get_path(target_work_dir_path, 'run_nltha.py'), 'w') as fid:
            fid.writelines(lines)

    def get_prelim_analysis_files(self):
        return ['generate_damping.py', 'model_param_vals.json', 'model_attributes.json']

    def get_nltha_files(self):
        return ['evaluate_edps.py', 'generate_edp_recorders.py']

    def run_prelim_analysis(self, target_work_dir_path, comp_env):
        self.run_file(target_work_dir_path, 'run_prelim_analysis.py', comp_env)

    def run_nltha(self, target_work_dir_path, comp_env):
        self.run_file(target_work_dir_path, 'run_nltha.py', comp_env)

    def run_file(self, target_work_dir_path, run_file_name, comp_env):
        curr_dir = Utility.get_path(os.getcwd())
        os.chdir(target_work_dir_path)
        if comp_env == 'local' and platform.system() == 'Windows':
            subprocess.call(
                [self.local_python_path, run_file_name], shell=True
            )
        os.chdir(curr_dir)

    def perform_mat_test(self, mat_data_dir_path, mat_tag, input_data, num_incr):
        raise NotImplementedError("openseespy implementation pending")

    @staticmethod
    def write_model_param_vals(work_dir_path, name_value_pairs):
        OpenSeesPy.write_json_input_params_file(work_dir_path, name_value_pairs, "model_param_vals.json")

    @staticmethod
    def write_model_attributes(work_dir_path, name_value_pairs):
        OpenSeesPy.write_json_input_params_file(work_dir_path, name_value_pairs, "model_attributes.json")

    @staticmethod
    def write_damping_file(damping_model, work_dir_path, **kwargs):
        lines = [
                f'import openseespy.opensees as ops\n',
                f'\n',
                f'\n',
                f'def generate_damping():\n',
        ]
        if damping_model == 'rayleigh_damping':
            alpha = kwargs['alpha']
            beta = kwargs['beta']
            lines.extend([
                f'    ops.rayleigh({alpha}, 0., 0., {beta})\n',
            ])
        if damping_model == 'modal_damping':
            n_modes = kwargs['n_modes']
            damp_ratios = kwargs['damp_ratios']
            lines.extend([
                f'    ops.eigen({int(n_modes)})\n',
                f"    ops.modalDamping({', '.join(np.array(damp_ratios).astype(str))})\n",
                f'    ops.wipeAnalysis()\n',
            ])
        if damping_model is None:
            lines.append('    pass\n')
        with open(Utility.get_path(work_dir_path, 'generate_damping.py'), 'w') as fid:
            fid.writelines(lines)
        return

    @staticmethod
    def get_stress_strain_recorder_command(recorder_file_path, ele_tag, sec_num, y, z, mat_tag):
        new_recorder_line = f'    ops.recorder(\'Element\', \'-file\', \'{recorder_file_path}\', \'-time\', \'-ele\', {ele_tag}, ' \
                            f'\'section\', {sec_num}, \'fiber\', {y}, {z}, {mat_tag}, \'stressStrain\')\n'
        return new_recorder_line

    @staticmethod
    def get_elem_deformation_recorder_command(recorder_file_path, ele_tag):
        new_recorder_line = f'    ops.recorder(\'Element\', \'-file\', \'{recorder_file_path}\', \'-time\', \'-ele\', {ele_tag}, \'deformation\')\n'
        return new_recorder_line

    @staticmethod
    def get_recorder_files_to_read(rec_gen_file_path):
        with open(rec_gen_file_path, 'r') as fid:
            lines = fid.readlines()
        lines = [[item.strip(' ,') for item in line.split(',')] for line in lines if 'ops.recorder(' in line and '-file' in line]
        return [line[line.index('\'-file\'') + 1].strip('\'') for line in lines]

    @staticmethod
    def import_module(module_path, module_name, alias, tab=''):
        # Assumes importlib.util is imported
        module_path = Utility.get_path(module_path, module_name + '.py')
        lines = [
            f"{tab}spec = importlib.util.spec_from_file_location('{module_name}', '{module_path}')\n",
            f"{tab}{alias} = importlib.util.module_from_spec(spec)\n",
            f"{tab}spec.loader.exec_module({alias})\n",
        ]
        return lines

    @staticmethod
    def read_json(file_path, var_name):
        # Assumes json is imported
        lines = [
            f'with open(\'{file_path}\', \'r\') as file:\n',
            f'    {var_name} = json.load(file)\n',
        ]
        return lines

    @staticmethod
    def create_dict(names, values, dict_name):
        lines = [f'{dict_name} = {{\n']
        for name, value in zip(names, values):
            lines.append(f'    \'{name}\': {value},\n')
        lines.append('}\n')
        return lines

    @staticmethod
    def write_json_input_params_file(work_dir_path, name_value_pairs, filename):
        to_dump_names = []
        to_dump_values = []
        for name, value in name_value_pairs:
            to_dump_names.extend(name)
            to_dump_values.extend(value)
        to_dump = dict(zip(to_dump_names, to_dump_values))
        with open(Utility.get_path(work_dir_path, filename), 'w') as fid:
            json.dump(to_dump, fid, indent=4)

    @staticmethod
    def unique_append(new_commands, lines, ex_commands):
        for command in new_commands:
            if command not in ex_commands:
                lines.append(command)
        return lines



# -*- coding: utf-8 -*-
"""
Created on Sun Nov 17 16:16:25 2019

@author: Angshuman Deb
"""

import numpy as np
from .sa import Sa
from .utility import Utility
import os
import sys


class SaT(Sa):
    large_value = 1e99
    small_value = 1e-99

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, structure, gmm, correl_func, period):
        self.period = period
        super().__init__(structure, gmm, correl_func)

    ####################################################################################################################
    # Public Functionalities (instance methods)
    ####################################################################################################################

    def evaluate_period(self, for_which):
        period = self.period
        if isinstance(period, str) and period.startswith('T_'):
            structure = self.structure
            eval_period = structure.get_period_range_extremes(for_which, [period, ])[0]
        elif isinstance(period, float):
            eval_period = period
        else:
            return

        rec_data_file_path = Utility.get_path(os.path.dirname(__file__), 'Rec_Data', 'rec_data.pickle')
        rec_data = Utility.pickle_load_dict(rec_data_file_path)
        per_known = rec_data['per_known']
        return per_known[np.argmin(np.abs(per_known - eval_period))]

    # ------------------------------------------------------------------------------------------------------------------

    def setup_psa_spectra_gms(self, mrp, for_which, mean_req_exact, sig_req_exact, spectral_periods,
                              gm, sa_known, per_known, is_scaled):
        target = self.get_inv_seismic_hazard(mrp, for_which)
        period = self.evaluate_period(for_which)
        period_mask_lt = spectral_periods < period
        period_mask_gt = spectral_periods > period

        mean_req_exact = np.hstack(
            (mean_req_exact[period_mask_lt],
             np.log(target),
             mean_req_exact[period_mask_gt])
        )

        sig_req_exact = np.hstack(
            (sig_req_exact[period_mask_lt],
             0.0,
             sig_req_exact[period_mask_gt])
        )

        gm = np.column_stack(
            (gm[:, period_mask_lt],
             target * np.ones(gm.shape[0]),
             gm[:, period_mask_gt])
        )

        spectral_periods = Utility.merge_sorted_1d_arrays(spectral_periods, [period])
        # Here zeros are ok as intialized values of indices because they are going to be filled up before they are used
        rec_per = np.zeros(len(spectral_periods)).astype(int)
        for i in range(len(spectral_periods)):
            rec_per[i] = np.argmin(np.abs(per_known - spectral_periods[i]))

        sa_known[sa_known == 0] = sys.float_info.epsilon
        sample_big = np.log(sa_known[:, rec_per])

        ind = np.in1d(spectral_periods, period)
        rec_val = np.exp(sample_big[:, ind]).flatten()
        if is_scaled:
            scale_fac = target / rec_val
        else:
            scale_fac = rec_val / rec_val

        return sample_big, scale_fac, mean_req_exact, sig_req_exact, gm

    # ------------------------------------------------------------------------------------------------------------------

    def evaluate_gmm(self, period):
        med_sa_period, sig_ln_sa_period = self._evaluate_gmm(period)
        # return 1-d arrays
        return med_sa_period.flatten(), sig_ln_sa_period.flatten()

    # ------------------------------------------------------------------------------------------------------------------

    def _get_target_spectrum(self, mrp, for_which, med_sa_spectrum, sig_ln_sa_spectrum,
                             spectral_periods, uncondition):
        if uncondition:
            scenario_weights = self.get_seismic_hazard_deagg(mrp, for_which)
            mean_ln_sa_spectrum, sig_ln_sa_spectrum = Utility.get_exact_mixture_params(
                np.log(med_sa_spectrum),
                sig_ln_sa_spectrum,
                scenario_weights)  # (1, n_sp), (1, n_sp)
            med_sa_spectrum = np.exp(mean_ln_sa_spectrum)  # (1, n_sp)

        period = self.evaluate_period(for_which)

        ind_period = np.in1d(spectral_periods, period)
        med_sa_period = med_sa_spectrum[:, ind_period].flatten()  # (n_s,)
        sig_ln_sa_period = sig_ln_sa_spectrum[:, ind_period].flatten()  # (n_s,)

        n_s = med_sa_spectrum.shape[0]
        n_sp = len(spectral_periods)

        rho = np.zeros((n_s, n_sp))
        rho[0:, :] = self.correl_func(spectral_periods, period)  # (n_s, n_sp)

        mean_req, cov_req = self.get_conditional_spectrum(mrp, for_which,
                                                          med_sa_spectrum,
                                                          sig_ln_sa_spectrum,
                                                          rho, self.get_moc(spectral_periods),
                                                          med_sa_period, sig_ln_sa_period)

        cov_req[ind_period, :, :] = 0.
        cov_req[:, ind_period, :] = 0.

        return mean_req, cov_req, spectral_periods

    # ------------------------------------------------------------------------------------------------------------------

    def mark_period_on_axis(self, for_which, ax, **kwargs):
        lc = kwargs.get('period_lc', 'black')
        lw = kwargs.get('period_lw', 1)
        period = self.evaluate_period(for_which)
        ylim = Utility.get_ylim(ax, (0.01, None))
        if '3D' in ax.__class__.__name__:
            ax.plot([period] * 2, ylim, 0, color=lc, linestyle='-', linewidth=lw)
        else:
            ax.plot([period] * 2, ylim, color=lc, linestyle='-', linewidth=lw)
        ax.set_ylim(ylim)
